import logging
import sys
import time

import requests


class AuthClient:
    _token = None
    credentials = None
    token_expires_at = None
    token_expiry_seconds = 1 * 60 * 60  # token expires after 1 hour

    def __init__(self, url, client_secret, standard_headers):
        self.url = url
        self.client_secret = client_secret
        self.standard_headers = standard_headers

    @property
    def token(self):
        if self._token is None or time.time() >= self.token_expires_at:

            headers = {'Content-Type': 'application/x-www-form-urlencoded'}
            headers.update(self.standard_headers)

            data = {"grant_type": "password",
                    "client_id": "f20cd2d3-682a-4568-a53e-4262ef54c8f4",
                    "scope": "openid",
                    "client_secret": self.client_secret,
                    "username": self.credentials.username,
                    "password": self.credentials.password
                    }

            try:
                r = requests.post(self.url, headers=headers, data=data)
                logging.info('')
                r.raise_for_status()
                reply = r.json()
                oauth_token = reply['access_token']
                logging.info(f"Authentication success for user '{self.credentials.username}'")
            except ConnectionError:
                logging.exception(f"Could not connect to the authentication service at {self.url}. "
                                  f"Check that the necessary outbound ports are open in your firewall. "
                                  f"See the documentation for more information.")
                sys.exit()
            except Exception:
                logging.exception(
                    "Invalid username, password or secret key - please check and retry."
                    " If problem persists contact helpdesk on helpdesk@ega-archive.org")
                sys.exit()

            self._token = oauth_token
            self.token_expires_at = time.time() + self.token_expiry_seconds

        return self._token



import os
import re
import shutil

from tests.functional.util import run

script_dir = os.path.dirname(__file__)


def test_download_file():
    file = 'EGAF00001753741'
    download_dir = f'{script_dir}/{file}'

    _assert_successful_run(f'pyega3 -t fetch {file} --output-dir {script_dir}', download_dir)
    _assert_complete_files(download_dir)
    _cleanup(download_dir)


def test_multipart_downloading():
    file = 'EGAF00005001625'  # less 200MB, will create 2 slices
    conns = 2  # only 2 will be utilised
    download_dir = f'{script_dir}/{file}'

    _assert_successful_run(f'pyega3 -t -c {conns} fetch {file} --output-dir {script_dir}', download_dir)
    _assert_complete_files(download_dir)
    _cleanup(download_dir)


def _assert_successful_run(command: str, download_dir: str):
    exit_code, output, error = run(command)
    assert exit_code == 0
    output += error  # it seems that output is in stderr
    assert bool(re.search(f'Download complete', output))


def _assert_complete_files(download_dir):
    downloaded_files = [f for f in os.listdir(download_dir) if os.path.isfile(f'{download_dir}/{f}')]
    assert len(downloaded_files) == 2


def _cleanup(download_dir):
    shutil.rmtree(download_dir)
    os.remove('pyega3_output.log')



import json

import pytest

from pyega3.libs.server_config import ServerConfig


def test_load_server_config_invalid_path():
    with pytest.raises(SystemExit):
        ServerConfig.from_file("/invalidpath");


def test_load_server_config_invalid_json(mock_input_file):
    with mock_input_file("bad json") as server_config_file:
        with pytest.raises(SystemExit):
            ServerConfig.from_file(server_config_file)


def test_load_server_config_missing_attributes_in_json_file(mock_input_file):
    config = {"url_auth": "http://url_auth", "url_api": "http://url_api"}

    with mock_input_file(json.dumps(config)) as server_config_file:
        with pytest.raises(SystemExit):
            ServerConfig.from_file(server_config_file)



import os
import re

import pytest
import responses

from pyega3.libs.data_file import DOWNLOAD_FILE_MEMORY_BUFFER_SIZE, DataFile

test_file_id = 'test_file_id1'
expected_file_size = DOWNLOAD_FILE_MEMORY_BUFFER_SIZE * 3


def test_temp_files_are_deleted_automatically_if_there_are_no_exceptions(mock_server_config,
                                                                         mock_auth_client,
                                                                         temporary_output_file,
                                                                         mock_requests,
                                                                         mock_data_client):
    """
    The temporary files are deleted by the algorithm automatically, during the happy path,
    when the temporary files are assembled into the final, big file.
    There's no need for extra deleting-mechanism.
    """
    DataFile.temporary_files_should_be_deleted = False

    file_size_without_iv = 92700
    file_size_with_iv = file_size_without_iv + 16

    input_file = bytearray(os.urandom(file_size_without_iv))
    mock_requests.add(responses.GET, f'{mock_server_config.url_api}/files/{test_file_id}', body=input_file, status=200)

    file = DataFile(mock_data_client, test_file_id, temporary_output_file, temporary_output_file, file_size_with_iv,
                    'check_sum')

    output_dir = os.path.dirname(temporary_output_file)
    file.download_file_retry(1, output_dir, None, 2, 0.1)

    temp_file = file.temporary_files.pop()
    # The temporary file should not exist because everything went fine,
    # and it was deleted automatically:
    assert not os.path.exists(temp_file)

    display_file_name = os.path.basename(temporary_output_file)
    downloaded_file = f'{output_dir}/{test_file_id}/{display_file_name}'
    assert os.path.exists(downloaded_file)
    output_file_size = os.stat(downloaded_file).st_size
    assert output_file_size == file_size_without_iv
    os.remove(downloaded_file)


def download_with_exception(mock_requests, output_file_path, mock_server_config, file):
    """
    Simulates downloading a file of the given size: "true_file_size".
    During the transfer, an exception happens and the temporary file is either deleted
    or kept, depending on the TEMPORARY_FILES_SHOULD_BE_DELETED flag.
    """

    number_of_retries = 2
    not_enough_bytes = int(expected_file_size / 3 - 1000)
    content = bytearray(os.urandom(not_enough_bytes))
    output_dir = os.path.dirname(output_file_path)

    # First, normal GET request:
    mock_requests.add(responses.GET, f'{mock_server_config.url_api}/files/{file.id}', body=content, status=200)
    # Then all the retry attempt:
    for _ in range(number_of_retries):
        mock_requests.add(responses.GET, f'{mock_server_config.url_api}/files/{file.id}', body=content, status=200)

    with pytest.raises(Exception) as context_manager:
        file.download_file_retry(1, output_dir, None, number_of_retries, 0.1)

    exception_message = str(context_manager.value)
    assert re.compile(r'Slice error: received=\d+, requested=\d+').search(exception_message)

    display_file_name = os.path.basename(output_file_path)
    downloaded_file = f'{output_dir}/{test_file_id}/{display_file_name}'
    assert not os.path.exists(downloaded_file)


def test_temporary_folder_is_deleted_if_the_user_says_so(mock_server_config,
                                                         mock_data_client,
                                                         temporary_output_file,
                                                         mock_requests):
    # Given: a file that exist in EGA object store and the user has permissions to access to it
    DataFile.temporary_files_should_be_deleted = True

    file = DataFile(mock_data_client, test_file_id, temporary_output_file, temporary_output_file, expected_file_size,
                    'check_sum')

    temporary_folder_name = os.path.join(os.path.dirname(temporary_output_file), test_file_id, '.tmp_download')

    # When: the user completes downloading a file
    download_with_exception(mock_requests, temporary_output_file, mock_server_config, file)

    # Then: the temporary folder and the temporary files are deleted
    assert not os.path.exists(temporary_folder_name)


def test_temporary_folder_is_not_deleted_if_the_user_says_so(mock_server_config,
                                                             mock_data_client,
                                                             temporary_output_file,
                                                             mock_requests):

    # Given: a file that exist in EGA object store and the user has permissions to access to it
    DataFile.temporary_files_should_be_deleted = False

    file = DataFile(mock_data_client, test_file_id, temporary_output_file, temporary_output_file, expected_file_size,
                    'check_sum')

    temporary_folder_name = os.path.join(os.path.dirname(temporary_output_file), test_file_id, '.tmp_download')

    # When: he user completes downloading a file
    download_with_exception(mock_requests, temporary_output_file, mock_server_config, file)

    # Then: the temporary folder and the temporary files are NOT deleted
    assert os.path.exists(temporary_folder_name)



import responses

from pyega3.libs.commands import execute_subcommand


def test_error_5xx(mock_data_client, mock_requests, mock_server_config, caplog):

    class Args:
        subcommand = "datasets"

    mock_requests.add(responses.GET, mock_server_config.url_api + "/metadata/datasets", status=503)

    try:
        execute_subcommand(Args(), mock_data_client)
    except SystemExit as sys_exit_error:
        assert sys_exit_error.code == 1

    errors = [m for m in caplog.records if m.levelname == "ERROR"]
    assert len(errors) == 1
    assert "error on the server" in errors[0].message


def test_error_too_many_requests(mock_data_client, mock_requests, mock_server_config, caplog):

    class Args:
        subcommand = "datasets"

    mock_requests.add(responses.GET, mock_server_config.url_api + "/metadata/datasets", status=429)

    try:
        execute_subcommand(Args(), mock_data_client)
    except SystemExit as sys_exit_error:
        assert sys_exit_error.code == 1

    errors = [m for m in caplog.records if m.levelname == "ERROR"]
    assert len(errors) == 1
    assert "too many requests" in errors[0].message



import hashlib
import logging
import os

import pytest

from pyega3.libs import utils
from pyega3.libs.data_file import DataFile

FILE_ID = 'EGAF00000000001'
DISPLAY_FILE_NAME = 'ENCFF284YOU.bam.bai'


@pytest.fixture()
def file_in_fire(mock_data_server, mock_data_client, random_binary_file, fs):
    mock_data_server.file_content[FILE_ID] = random_binary_file
    file_md5 = hashlib.md5(random_binary_file).hexdigest()
    yield DataFile(mock_data_client, FILE_ID, display_file_name=DISPLAY_FILE_NAME,
                   file_name=DISPLAY_FILE_NAME + '.cip', size=len(random_binary_file) + 16,
                   unencrypted_checksum=file_md5)


def download_file_from_fire(file_in_fire, output_dir):
    # num_connections, output_dir, genomic_range_args, max_retries, retry_wait
    file_in_fire.download_file_retry(1, output_dir, None, 2, 0)


def test_file_is_saved_into_an_existing_directory_which_was_specified_by_the_user(file_in_fire):
    """
    The user wants to download a file into a specific, user-specified directory.
    The directory exists and there is no such file in that directory yet.
    Expected: the file is downloaded into the user-specified directory
    and the file-name will be the 'displayFileName' (e.g. 'EGAF00000000001/ENCFF284YOU.bam.bai').
    """

    output_dir = 'user_specified_directory'
    os.makedirs(output_dir)

    expected_local_file_name = f'/{output_dir}/{FILE_ID}/{DISPLAY_FILE_NAME}'
    expected_local_md5_file_name = f'{expected_local_file_name}.md5'

    assert os.path.isdir(output_dir)
    assert not os.path.exists(expected_local_file_name)
    assert not os.path.exists(expected_local_md5_file_name)

    download_file_from_fire(file_in_fire, output_dir=output_dir)

    assert os.path.isfile(expected_local_file_name)
    assert os.path.isfile(expected_local_md5_file_name)


def test_file_in_directory_is_not_downloaded_again(file_in_fire, random_binary_file, caplog, fs):
    """
    When the file has correctly been downloaded already into the specified directory,
    then it is not downloaded again.
    """
    caplog.set_level(logging.INFO)
    output_dir = 'user_specified_directory_name'
    expected_local_file_name = f'/{output_dir}/{FILE_ID}/{DISPLAY_FILE_NAME}'
    expected_local_md5_file_name = f'{expected_local_file_name}.md5'

    # First I create the downloaded files to simulate a previous, successful download:
    fs.create_file(expected_local_file_name)
    fs.create_file(expected_local_md5_file_name)
    assert os.path.isdir(output_dir)

    with open(expected_local_file_name, 'wb') as fh:
        fh.write(random_binary_file)
    with open(expected_local_md5_file_name, 'w') as fh:
        fh.write(file_in_fire.unencrypted_checksum)
    # And I store their original mtimes:
    local_file_orig_mtime = os.stat(expected_local_file_name).st_mtime
    local_md5_file_orig_mtime = os.stat(expected_local_md5_file_name).st_mtime

    # The downloaded files should exist:
    assert os.path.isdir(output_dir)
    assert os.path.isfile(expected_local_file_name)
    assert os.path.isfile(expected_local_md5_file_name)

    # Let's simulate a second download into the specified directory.
    # This should not download again the files,
    # but it should print out a "Local file exists" message.
    download_file_from_fire(file_in_fire, output_dir=output_dir)

    local_file_mtime_after_download = os.stat(expected_local_file_name).st_mtime
    local_md5_file_mtime_after_download = os.stat(expected_local_md5_file_name).st_mtime

    # If the files haven't been re-downloaded again, then the mtimes should not have changed:
    assert local_file_orig_mtime == local_file_mtime_after_download
    assert local_md5_file_orig_mtime == local_md5_file_mtime_after_download

    assert f"Local file exists:'{expected_local_file_name}'" in caplog.text


def test_corrupted_file_in_a_directory_is_downloaded_again(file_in_fire, fs):
    """
    When the file has been downloaded incorrectly into a directory previously, then it is re-downloaded again.
    I don't think this happens too often, because an incorrectly downloaded file is always removed
    (or at least, it should be always removed).
    """

    output_dir = 'user_specified_directory_name'
    expected_local_file_name = f'/{output_dir}/{FILE_ID}/{DISPLAY_FILE_NAME}'
    expected_local_md5_file_name = f'{expected_local_file_name}.md5'

    # First I create a file to simulate a previous, unsuccessful download:
    fs.create_file(expected_local_file_name)
    with open(expected_local_file_name, 'w') as fh:
        fh.write('incorrect, corrupted content')

    # The downloaded file should exist, but without an .md5:
    assert os.path.isfile(expected_local_file_name)
    assert not os.path.exists(expected_local_md5_file_name)
    assert os.path.isdir(output_dir)

    # Let's simulate a second download into the specified directory.
    # This should download again the file and the .md5, this time, correctly.
    download_file_from_fire(file_in_fire, output_dir=output_dir)

    # Check that the actual MD5 of the re-downloaded file is correct:
    correct_expected_md5 = file_in_fire.unencrypted_checksum
    actual_md5 = utils.calculate_md5(expected_local_file_name, file_in_fire.size)
    assert actual_md5 == correct_expected_md5

    # The .md5 file should now exist:
    assert os.path.isfile(expected_local_md5_file_name)

    # And the content of the .md5 file should be the correct, expected one:
    with open(expected_local_md5_file_name, 'r') as fh:
        actual_md5 = fh.readline()
    assert actual_md5 == correct_expected_md5


def test_non_existing_directory_is_created(file_in_fire):
    """
    The user wants to download a file into a specific, user-specified directory.
    The directory does not exist.
    Expected: the non-existent directory is created automatically for the user,
    the file is downloaded into the user-specified directory
    and the file-name will be the 'displayFileName' (e.g. 'EGAF00000000001/ENCFF284YOU.bam.bai').
    The non-existent directory is created automatically as a (positive) side-effect of the code
    which creates the .tmp_download directory: that code creates not only the .tmp_download directory,
    but all the missing parent directories, too, including the non-existent user-specified directory.
    """

    output_dir = 'user_specified_directory'
    expected_local_file_name = f'/{output_dir}/{FILE_ID}/{DISPLAY_FILE_NAME}'
    expected_local_md5_file_name = f'{expected_local_file_name}.md5'

    assert not os.path.exists(output_dir)
    assert not os.path.exists(expected_local_file_name)
    assert not os.path.exists(expected_local_md5_file_name)

    download_file_from_fire(file_in_fire, output_dir=output_dir)

    assert os.path.isdir(output_dir)
    assert os.path.isfile(expected_local_file_name)
    assert os.path.isfile(expected_local_md5_file_name)



import json
from urllib import parse

import pytest
import responses

import tests.unit.conftest as common
from pyega3.libs.auth_client import AuthClient
from pyega3.libs.credentials import Credentials


@pytest.fixture
def mock_openid_server(mock_requests, mock_server_config):
    class MockOpenIDServer:
        url = mock_server_config.url_auth
        id_token = common.rand_str()
        access_token = common.rand_str()
        username = common.rand_str()
        password = common.rand_str()

        def __init__(self):
            mock_requests.add_callback(responses.POST,
                                       self.url,
                                       callback=self.request_callback,
                                       content_type='application/json')

        def request_callback(self, request):

            query = parse.parse_qs(request.body)
            if query['username'][0] == self.username and query['password'][0] == self.password:
                return (200, {}, json.dumps(
                    {"access_token": self.access_token, "id_token": self.id_token, "token_type": "Bearer",
                     "expires_in": 3600}))
            else:
                return 400, {}, json.dumps({"error_description": "Bad credentials", "error": "invalid_grant"})

    return MockOpenIDServer()


def test_get_token_from_openid_server(mock_openid_server, mock_server_config):
    good_credentials = Credentials(username=mock_openid_server.username, password=mock_openid_server.password)
    auth_server = AuthClient(mock_openid_server.url, mock_server_config.client_secret, {})
    auth_server.credentials = good_credentials
    assert auth_server.token == mock_openid_server.access_token


def test_bad_openid_credentials_exits(mock_openid_server, mock_server_config):
    bad_credentials = Credentials(username=common.rand_str(), password=common.rand_str())
    auth_server = AuthClient(mock_openid_server.url, mock_server_config.client_secret, {})
    auth_server.credentials = bad_credentials
    with pytest.raises(SystemExit):
        token = auth_server.token



import logging
import json

from pyega3.libs.utils import status_ok


def pretty_print_authorized_datasets(datasets, as_json):
    if as_json:
        logging.info(json.dumps([{"id": dataset.id for dataset in datasets}]))
    else:
        logging.info("Dataset ID")
        logging.info("-----------------")
        for dataset in datasets:
            logging.info(dataset.id)


def pretty_print_files_in_dataset(files, as_json):
    """
    Print a table of files in authorized dataset from api call api_list_files_in_dataset

        {
           "checksumType": "MD5",
            "unencryptedChecksum": "MD5SUM678901234567890123456789012",
            "fileName": "EGAZ00001314035/b37/NA12878.bam.bai.cip",
            "displayFileName": "NA12878.bam.bai.cip",
            "fileStatus": "available",
            "fileSize": 8949984,
            "datasetId": "EGAD00001003338",
            "fileId": "EGAF00001753747"
        }

    """

    if as_json:
        logging.info(json.dumps([{
            "id": file.id,
            "status": file.status,
            "bytes": file.size,
            "checksum": file.unencrypted_checksum,
            "name": file.display_name
        } for file in files]))
    else:
        format_string = "{:15} {:6} {:12} {:36} {}"

        logging.info(format_string.format("File ID", "Status", "Bytes", "Check sum", "File name"))
        for file in files:
            logging.info(format_string.format(file.id, status_ok(file.status), str(file.size),
                                              file.unencrypted_checksum, file.display_name))

        logging.info('-' * 80)
        logging.info("Total dataset size = %.2f GB " % (sum(file.size for file in files) / (1024 * 1024 * 1024.0)))



import contextlib
import json
import logging

import requests
from requests import Session
from requests.adapters import HTTPAdapter, DEFAULT_POOLSIZE
from urllib3.util import retry


def create_session_with_retry(retry_policy: retry.Retry = None, pool_max_size=None) -> Session:
    retry_policy = retry_policy or retry.Retry(
        status_forcelist=[429, 500, 503, 504],
        # status is the no. of retries if response is in status_forcelist
        status=10,
        # total has a default value of 10,
        # need to set total to a number higher than status so it'll respect the status retry count
        total=20,
        # do not retry connection errors
        connect=False,
        read=10,
        # 0.3, 0.6, 1.2, 2.4, 4.8, 9.6, 19.2, 38.4, 76.8, 120 is the BACKOFF_MAX in Retry
        backoff_factor=0.6
    )
    session = Session()
    POOL_MAX_SIZE = max(DEFAULT_POOLSIZE, pool_max_size or 0)
    adapter = HTTPAdapter(max_retries=retry_policy, pool_maxsize=POOL_MAX_SIZE)
    session.mount('https://', adapter)
    return session


class DataClient:

    def __init__(self, url, htsget_url, auth_client, standard_headers, connections=None):
        self.url = url
        self.htsget_url = htsget_url
        self.auth_client = auth_client
        self.standard_headers = standard_headers
        self.session = create_session_with_retry(pool_max_size=connections)

    @staticmethod
    def print_debug_info(url, reply_json, *args):
        logging.debug(f"Request URL : {url}")
        if reply_json is not None:
            logging.debug("Response    :\n %.1200s" % json.dumps(reply_json, indent=4))

        for a in args:
            logging.debug(a)

    def get_json(self, path):
        headers = {'Accept': 'application/json', 'Authorization': f'Bearer {self.auth_client.token}'}
        headers.update(self.standard_headers)

        url = f"{self.url}{path}"
        r = self.session.get(url, headers=headers)
        r.raise_for_status()

        reply = r.json()

        self.print_debug_info(url, reply)
        return reply

    @contextlib.contextmanager
    def get_stream(self, path, extra_headers=None):
        headers = {'Authorization': f'Bearer {self.auth_client.token}'}
        headers.update(self.standard_headers)
        if extra_headers is not None:
            headers.update(extra_headers)

        url = f'{self.url}{path}'
        with self.session.get(url, headers=headers, stream=True) as r:
            self.print_debug_info(url, None, f"Response headers: {r.headers}")
            r.raise_for_status()
            yield r



import logging
import sys

from pyega3.libs import data_file
from pyega3.libs.data_file import DataFile
from pyega3.libs.utils import status_ok

LEGACY_DATASETS = ["EGAD00000000003", "EGAD00000000004", "EGAD00000000005", "EGAD00000000006", "EGAD00000000007",
                   "EGAD00000000008", "EGAD00000000009", "EGAD00000000025", "EGAD00000000029", "EGAD00000000043",
                   "EGAD00000000048", "EGAD00000000049", "EGAD00000000051", "EGAD00000000052", "EGAD00000000053",
                   "EGAD00000000054", "EGAD00000000055", "EGAD00000000056", "EGAD00000000057", "EGAD00000000060",
                   "EGAD00000000114", "EGAD00000000119", "EGAD00000000120", "EGAD00000000121", "EGAD00000000122",
                   "EGAD00001000132", "EGAD00010000124", "EGAD00010000144", "EGAD00010000148", "EGAD00010000150",
                   "EGAD00010000158", "EGAD00010000160", "EGAD00010000162", "EGAD00010000164", "EGAD00010000246",
                   "EGAD00010000248", "EGAD00010000250", "EGAD00010000256", "EGAD00010000444"]


class DataSet:
    def __init__(self, data_client, dataset_id):
        self.id = dataset_id
        self.data_client = data_client

    @staticmethod
    def list_authorized_datasets(data_client):
        """List datasets to which the credentialed user has authorized access"""

        reply = data_client.get_json("/metadata/datasets")

        if reply is None:
            logging.error(
                "You do not currently have access to any datasets at EGA according to our databases."
                " If you believe you should have access please contact helpdesk on helpdesk@ega-archive.org")
            sys.exit()

        return [DataSet(data_client, dataset_id) for dataset_id in reply]

    def list_files(self):
        if self.id in LEGACY_DATASETS:
            logging.error(f"This is a legacy dataset {self.id}. Please contact the EGA helpdesk at "
                          f"helpdesk@ega-archive.org for more information.")
            sys.exit()

        authorized_datasets = DataSet.list_authorized_datasets(self.data_client)

        if self.id not in [dataset.id for dataset in authorized_datasets]:
            logging.error(f"Dataset '{self.id}' is not in the list of your authorized datasets.")
            sys.exit()

        reply = self.data_client.get_json(f"/metadata/datasets/{self.id}/files")

        if reply is None:
            logging.error(f"List files in dataset {self.id} failed")
            sys.exit()

        def make_data_file(res):
            display_file_name = res['displayFileName'] if 'displayFileName' in res else None
            file_name = res['fileName'] if 'fileName' in res else None
            size = res['fileSize'] if 'fileSize' in res else None
            unencrypted_checksum = res['unencryptedChecksum'] if 'unencryptedChecksum' in res else None
            return data_file.DataFile(self.data_client, res['fileId'],
                                      display_file_name=display_file_name,
                                      file_name=file_name,
                                      size=size,
                                      unencrypted_checksum=unencrypted_checksum,
                                      status=res['fileStatus'])

        return [make_data_file(res) for res in reply]

    def download(self, num_connections, output_dir, genomic_range_args, max_retries=5, retry_wait=5,
                 max_slice_size=DataFile.DEFAULT_SLICE_SIZE):
        if self.id in LEGACY_DATASETS:
            logging.error(
                f"This is a legacy dataset {self.id}. Please contact the EGA helpdesk at helpdesk@ega-archive.org for more information.")
            sys.exit()

        authorized_datasets = DataSet.list_authorized_datasets(self.data_client)

        if self.id not in [dataset.id for dataset in authorized_datasets]:
            logging.info(f"Dataset '{self.id}' is not in the list of your authorized datasets.")
            return

        files = self.list_files()
        for file in files:
            try:
                if status_ok(file.status):
                    file.download_file_retry(num_connections, output_dir, genomic_range_args, max_retries, retry_wait, max_slice_size)
            except Exception as e:
                logging.exception(e)



import os

import pytest

from pyega3.libs.utils import verify_output_dir


def test_absolute_path_is_returned_if_directory_exists(fs):
    output_dir = 'user_specified_directory'
    os.makedirs(output_dir)

    assert os.path.exists(output_dir)
    assert os.path.isdir(output_dir)

    verified_output_dir = verify_output_dir(output_dir)

    absolute_path_of_output_dir = '/user_specified_directory'
    assert verified_output_dir == absolute_path_of_output_dir


def test_error_is_thrown_if_directory_does_not_exist(fs):
    """
    The user wants to download a file into a specific, user-specified directory,
    but that directory does not exist. In this case, an error is thrown.
    """

    output_dir = 'user_specified_directory'
    assert not os.path.exists(output_dir)

    with pytest.raises(NotADirectoryError) as exception_info:
        verify_output_dir(output_dir)

    assert exception_info.value.args[0] == 'The "user_specified_directory" directory, which was specified by ' \
                                           'the --output-dir command-line argument, is not an existing directory. ' \
                                           'Please either create that directory or specify a different one.'


def test_if_specified_dir_is_a_file_then_error_is_thrown(fs):
    """
    The user wants to download a file into a specific, user-specified directory,
    but that directory is, in fact, a file. An error is thrown and nothing is downloaded.
    """

    output_dir = 'user_specified_directory'
    existing_file = output_dir

    fs.create_file(existing_file)
    assert os.path.exists(existing_file)
    assert os.path.isfile(existing_file)

    with pytest.raises(NotADirectoryError) as exception_info:
        verify_output_dir(existing_file)

    assert exception_info.value.args[0] == 'The "user_specified_directory" directory, which was specified by ' \
                                           'the --output-dir command-line argument, is not an existing directory. ' \
                                           'Please either create that directory or specify a different one.'

    assert os.path.exists(existing_file)
    assert os.path.isfile(existing_file)



import logging
import logging.handlers
import sys

from requests.exceptions import HTTPError

from pyega3.libs import data_file
from pyega3.libs.data_set import DataSet
from pyega3.libs.pretty_printing import pretty_print_authorized_datasets, pretty_print_files_in_dataset
from pyega3.libs.utils import verify_output_dir


def execute_subcommand(args, data_client):
    try:
        if args.subcommand == "datasets":
            list_datasets(args, data_client)

        if args.subcommand == "files":
            list_files_in_dataset(args, data_client)

        elif args.subcommand == "fetch":
            fetch_data(args, data_client)
    except HTTPError as error:
        if error.response.status_code >= 500:
            logging.error("Could not communicate with the EGA repository because of an error on the server.\n"
                          "Please retry later, or contact helpdesk@ega-archive.org for assistance.\n"
                          f"(error detail: {error})")
        elif error.response.status_code == 429:
            logging.error("The EGA repository refused your request because you have made too many requests.\n"
                          "Please wait before making further requests. If you are running pyega3 in a script, please "
                          "add delays between your calls.")
        else:
            logging.error(f"There was a problem communicating with the EGA repository: {error}")

        sys.exit(1)


def fetch_data(args, data_client):
    output_dir = verify_output_dir(args.output_dir)
    genomic_range_args = (args.reference_name, args.reference_md5, args.start, args.end, args.format)
    if args.delete_temp_files:
        data_file.DataFile.temporary_files_should_be_deleted = True
    if args.identifier[3] == 'D':
        dataset = DataSet(data_client, args.identifier)
        dataset.download(args.connections, output_dir, genomic_range_args,
                         args.max_retries, args.retry_wait, args.max_slice_size)
    elif args.identifier[3] == 'F':
        file = data_file.DataFile(data_client, args.identifier)
        file.download_file_retry(num_connections=args.connections,
                                 output_dir=output_dir,
                                 genomic_range_args=genomic_range_args,
                                 max_retries=args.max_retries,
                                 retry_wait=args.retry_wait,
                                 max_slice_size=args.max_slice_size)
    else:
        logging.error(
            "Unrecognized identifier - please use EGAD accession for dataset request"
            " or EGAF accession for individual file requests")
        sys.exit()
    logging.info("Download complete")


def list_files_in_dataset(args, data_client):
    if args.identifier[3] != 'D':
        logging.error("Unrecognized identifier - please use EGAD accession for dataset requests")
        sys.exit()
    dataset = DataSet(data_client, args.identifier)
    files = dataset.list_files()
    pretty_print_files_in_dataset(files, args.json)


def list_datasets(args, data_client):
    datasets = DataSet.list_authorized_datasets(data_client)
    pretty_print_authorized_datasets(datasets, args.json)



from tests.functional.util import run


def test_list_datasets():
    exit_code, output, error = run(f'pyega3 -t datasets')
    assert exit_code == 0


def test_list_files():
    dataset = 'EGAD00001003338'
    exit_code, output, error = run(f'pyega3 -t files {dataset}')
    assert exit_code == 0



#!/usr/bin/env python3

import argparse
import logging
import logging.handlers
import os
import platform
import random

from pyega3.libs.auth_client import AuthClient
from pyega3.libs.credentials import Credentials
from pyega3.libs.data_client import DataClient
from pyega3.libs.server_config import ServerConfig
from pyega3.libs.utils import get_client_ip
from pyega3.libs.data_file import DataFile
from pyega3.libs.commands import execute_subcommand

VERSION = "4.0.4"
session_id = random.getrandbits(32)
logging_level = logging.INFO


def main():
    parser = argparse.ArgumentParser(description="Download from EMBL EBI's EGA (European Genome-phenome Archive)")
    parser.add_argument("-d", "--debug", action="store_true", help="Extra debugging messages")
    parser.add_argument("-cf", "--config-file", dest='config_file',
                        help='JSON file containing credentials/config e.g.{"username":"user1","password":"toor"}')
    parser.add_argument("-sf", "--server-file", dest='server_file',
                        help='JSON file containing server config e.g.{"url_auth":"aai url","url_api":"api url", '
                             '"url_api_ticket":"htsget url", "client_secret":"client secret"}')
    parser.add_argument("-c", "--connections", type=int, default=1,
                        help="Download using specified number of connections (default: 1 connection)")
    parser.add_argument("-t", "--test", action="store_true", help="Test user activated")
    parser.add_argument("-ms", "--max-slice-size", type=int, default=DataFile.DEFAULT_SLICE_SIZE,
                        help="Set maximum size for each slice in bytes (default: 100 MB)")
    parser.add_argument("-j", "--json", action="store_true", help="Output data in JSON format instead of tables")

    subparsers = parser.add_subparsers(dest="subcommand", help="subcommands")

    parser_ds = subparsers.add_parser("datasets", help="List authorized datasets")

    parser_dsinfo = subparsers.add_parser("files", help="List files in a specified dataset")
    parser_dsinfo.add_argument("identifier", help="Dataset ID (e.g. EGAD00000000001)")

    parser_fetch = subparsers.add_parser("fetch", help="Fetch a dataset or file")
    parser_fetch.add_argument("identifier",
                              help="Id for dataset (e.g. EGAD00000000001) or file (e.g. EGAF12345678901)")

    parser_fetch.add_argument(
        "--reference-name", "-r", type=str, default=None,
        help=(
            "The reference sequence name, for example 'chr1', '1', or 'chrX'. "
            "If unspecified, all data is returned."))
    parser_fetch.add_argument(
        "--reference-md5", "-m", type=str, default=None,
        help=(
            "The MD5 checksum uniquely representing the requested reference "
            "sequence as a lower-case hexadecimal string, calculated as the MD5 "
            "of the upper-case sequence excluding all whitespace characters."))
    parser_fetch.add_argument(
        "--start", "-s", type=int, default=None,
        help=(
            "The start position of the range on the reference, 0-based, inclusive. "
            "If specified, reference-name or reference-md5 must also be specified."))
    parser_fetch.add_argument(
        "--end", "-e", type=int, default=None,
        help=(
            "The end position of the range on the reference, 0-based exclusive. If "
            "specified, reference-name or reference-md5 must also be specified."))
    parser_fetch.add_argument(
        "--format", "-f", type=str, default=None, choices=["BAM", "CRAM"], help="The format of data to request.")

    parser_fetch.add_argument(
        "--max-retries", "-M", type=int, default=5,
        help="The maximum number of times to retry a failed transfer. Any negative number means infinite number of retries.")

    parser_fetch.add_argument(
        "--retry-wait", "-W", type=float, default=60,
        help="The number of seconds to wait before retrying a failed transfer.")

    parser_fetch.add_argument("--output-dir", default=os.getcwd(),
                              help="Output directory. The files will be saved into this directory. Must exist. "
                                   "Default: the current working directory.")

    parser_fetch.add_argument("--delete-temp-files", action="store_true",
                              help="Do not keep those temporary, partial files "
                                   "which were left on the disk after a failed transfer.")

    args = parser.parse_args()
    if args.debug:
        global logging_level
        logging_level = logging.DEBUG

    logging.basicConfig(level=logging_level,
                        format='%(asctime)s %(message)s',
                        datefmt='[%Y-%m-%d %H:%M:%S %z]',
                        handlers=[
                            logging.handlers.RotatingFileHandler("pyega3_output.log",
                                                                 maxBytes=5 * 1024 * 1024,
                                                                 backupCount=1),
                            logging.StreamHandler()
                        ])

    logging.info("")
    logging.info(f"pyEGA3 - EGA python client version {VERSION} (https://github.com/EGA-archive/ega-download-client)")
    logging.info(
        "Parts of this software are derived from pyEGA (https://github.com/blachlylab/pyega) by James Blachly")
    logging.info(f"Python version : {platform.python_version()}")
    logging.info(f"OS version : {platform.system()} {platform.version()}")
    if platform.system() == "Darwin":
        logging.info(f"MacOS version : {platform.mac_ver()[0]}")

    root_dir = os.path.split(os.path.realpath(__file__))[0]
    config_file_path = os.path.join(root_dir, "config", "default_credential_file.json")

    if args.test:
        credentials = Credentials.from_file(config_file_path)
    elif args.config_file is None:
        credentials = Credentials.from_file("credential_file.json")
    else:
        credentials = Credentials.from_file(args.config_file)

    if args.server_file is not None:
        server_config = ServerConfig.from_file(args.server_file)
    else:
        server_config = ServerConfig.from_file(ServerConfig.default_config_path())

    logging.info(f"Server URL: {server_config.url_api}")
    logging.info(f"Session-Id: {session_id}")

    standard_headers = {'Client-Version': VERSION, 'Session-Id': str(session_id), 'client-ip': get_client_ip()}

    auth_client = AuthClient(server_config.url_auth, server_config.client_secret, standard_headers)
    auth_client.credentials = credentials

    data_client = DataClient(server_config.url_api, server_config.url_api_ticket, auth_client, standard_headers,
                             connections=args.connections)

    execute_subcommand(args, data_client)


if __name__ == "__main__":
    main()



import json
import logging
import os
import sys


class ServerConfig:
    url_api = None
    url_auth = None
    url_api_ticket = None
    client_secret = None

    def __init__(self, url_api, url_auth, url_api_ticket, client_secret):
        self.url_api = url_api
        self.url_auth = url_auth
        self.url_api_ticket = url_api_ticket
        self.client_secret = client_secret

    @staticmethod
    def default_config_path():
        root_dir = os.path.split(os.path.realpath(__file__))[0]
        return os.path.join(root_dir, "../config", "default_server_file.json")

    @staticmethod
    def from_file(filepath):
        """Load server config for EMBL/EBI EGA from specified file"""
        filepath = os.path.expanduser(filepath)
        if not os.path.exists(filepath):
            logging.error(f"{filepath} does not exist")
            sys.exit()

        try:
            with open(filepath) as f:
                custom_server_config = json.load(f)

            def check_key(key):
                if key not in custom_server_config:
                    logging.error(f"{filepath} does not contain '{key}' field")
                    sys.exit()

            check_key('url_auth')
            check_key('url_api')
            check_key('url_api_ticket')
            check_key('client_secret')

            return ServerConfig(custom_server_config['url_api'],
                                custom_server_config['url_auth'],
                                custom_server_config['url_api_ticket'],
                                custom_server_config['client_secret'])

        except ValueError:
            logging.error("Invalid server config JSON file")
            sys.exit()



import contextlib
import hashlib
import os
import random
import string
import tempfile
from collections import namedtuple
from unittest import mock

import pytest
import responses
from psutil import virtual_memory

from pyega3.libs.data_client import DataClient
from pyega3.libs.server_config import ServerConfig
from tests.unit.mock_data_server import MockDataServer


def rand_str():
    length = random.randint(1, 127)
    return ''.join(random.choices(string.ascii_letters + string.digits, k=length))


@pytest.fixture
def mock_server_config():
    return ServerConfig(url_api='https://test.data.server',
                        url_auth='https://test.auth.server/ega-openid-connect-server/token',
                        url_api_ticket='https://test.ticket.server',
                        client_secret='test-client-secret')


@pytest.fixture
def user_has_authenticated_successfully(mock_requests, mock_server_config):
    mock_requests.add(responses.POST, mock_server_config.url_auth, json={'access_token': 'ok'}, status=200)


class MockAuthClient:
    credentials = None

    def __init__(self, token=None):
        if token is None:
            token = ''.join(random.choices(string.ascii_letters, k=64))
        self.token = token


@pytest.fixture
def mock_auth_client():
    return MockAuthClient()


@pytest.fixture
def mock_data_client(mock_server_config, mock_auth_client):
    return DataClient(mock_server_config.url_api, mock_server_config.url_api_ticket, mock_auth_client, {})


@pytest.fixture
def mock_requests():
    with responses.RequestsMock() as rsps:
        yield rsps


@pytest.fixture
def random_binary_file():
    mem = virtual_memory().available
    file_length = random.randint(1, mem // 512)
    return os.urandom(file_length)


another_random_binary_file = random_binary_file


@pytest.fixture
def temporary_output_file():
    """Returns a file-path to a random, temporary file-name."""
    _, output_file_path = tempfile.mkstemp()
    os.remove(output_file_path)
    return output_file_path


@pytest.fixture
def mock_data_server(mock_requests, mock_server_config, mock_auth_client):
    return MockDataServer(mock_requests, mock_server_config.url_api, mock_auth_client.token)


@pytest.fixture
def mock_input_file():
    @contextlib.contextmanager
    def make_mock_input_file(contents):
        file_name = rand_str()
        with mock.patch('os.path.exists', lambda p: p == file_name):
            with mock.patch('builtins.open', mock.mock_open(read_data=contents)):
                yield file_name

    return make_mock_input_file


@pytest.fixture
def empty_dataset(mock_data_server):
    file_id = "EGAD00000000001"
    mock_data_server.dataset_files[file_id] = None
    Dataset = namedtuple('Dataset', ['id'])
    return Dataset(file_id)


def mock_dataset_file(dataset_id, file_id, file_name, display_file_name, file_content):
    unencrypted_checksum = hashlib.md5(file_content).hexdigest()
    return {
        "unencryptedChecksum": unencrypted_checksum,
        "datasetId": dataset_id,
        "fileStatus": "available",
        "fileId": file_id,
        "checksumType": "MD5",
        "fileSize": len(file_content) + 16,
        "fileName": file_name,
        "displayFileName": display_file_name,
        "fileContent": file_content
    }


@pytest.fixture
def mock_dataset_files(empty_dataset, random_binary_file, another_random_binary_file):
    return [
        mock_dataset_file(empty_dataset.id, "EGAF00000000001",
                          "EGAZ00000000001/ENCFF000001.bam",
                          "ENCFF000001.bam", random_binary_file),
        mock_dataset_file(empty_dataset.id, "EGAF00000000002",
                          "EGAZ00000000002/ENCFF000002.bam",
                          "ENCFF000002.bam", another_random_binary_file)
    ]


@pytest.fixture
def dataset_with_files(mock_data_server, empty_dataset, mock_dataset_files):
    files = mock_dataset_files
    for file in files:
        # The file.pop("fileContent") call would change the original mock_dataset_file object,
        # which I avoid now by using copy():
        file = file.copy()
        file_id = file.get("fileId")
        mock_data_server.file_content[file_id] = file.pop("fileContent")
        mock_data_server.files[file_id] = file
    mock_data_server.dataset_files = {empty_dataset.id: [file.get("fileId") for file in files]}

    Dataset = namedtuple("DatasetWithFiles", ["id", "files"])
    return Dataset(empty_dataset.id, files)



import json

from pyega3.libs.data_file import DataFile
from pyega3.libs.data_set import DataSet
from pyega3.libs.pretty_printing import *


def test_pretty_print_authorized_datasets(mock_data_client):
    pretty_print_authorized_datasets([DataSet(mock_data_client, 'EGAD0123')], False)


def test_pretty_print_authorized_datasets_in_json(mock_data_client, caplog):
    caplog.set_level(logging.INFO)
    pretty_print_authorized_datasets([DataSet(mock_data_client, 'EGAD0123')], True)
    assert len(caplog.messages) == 1
    assert json.loads(caplog.messages[0]) == [{"id": "EGAD0123"}]


def test_pretty_print_files_in_dataset(mock_data_client):
    test_reply = [DataFile(mock_data_client, "EGAF00001753747",
                           display_file_name="EGAZ00001314035.bam.bai.cip",
                           file_name="EGAZ00001314035.bam.bai.cip",
                           size=0,
                           unencrypted_checksum="MD5SUM678901234567890123456789012",
                           status="available")]
    pretty_print_files_in_dataset(test_reply, False)


def test_pretty_print_files_in_dataset_in_json(mock_data_client, caplog):
    caplog.set_level(logging.INFO)
    test_reply = [DataFile(mock_data_client, "EGAF00001753747",
                           display_file_name="EGAZ00001314035.bam.bai.cip",
                           file_name="EGAZ00001314035.bam.bai.cip",
                           size=0,
                           unencrypted_checksum="MD5SUM678901234567890123456789012",
                           status="available")]
    pretty_print_files_in_dataset(test_reply, True)

    assert len(caplog.messages) == 1

    output_object = json.loads(caplog.messages[0])
    assert len(output_object) == 1
    assert output_object[0]["id"] == test_reply[0].id
    assert output_object[0]["name"] == test_reply[0].display_name
    assert output_object[0]["bytes"] == test_reply[0].size
    assert output_object[0]["checksum"] == test_reply[0].unencrypted_checksum
    assert output_object[0]["status"] == test_reply[0].status



import json
import re
import urllib

import responses


class MockDataServer:
    url = None
    token = None

    dataset_files = {}
    files = {}
    file_content = {}

    def __init__(self, mock_requests, url, token):
        self.url = url
        self.token = token

        mock_requests.add_callback(
            responses.GET,
            self.url + "/metadata/datasets",
            callback=self.list_datasets_callback,
            content_type='application/json',
        )

        mock_requests.add_callback(
            responses.GET,
            re.compile(self.url + "/metadata/datasets/.*/files"),
            callback=self.list_files_callback,
            content_type='application/json'
        )

        mock_requests.add_callback(
            responses.GET,
            re.compile(self.url + "/metadata/files/.*"),
            callback=self.get_file_metadata_callback,
            content_type='application/json'
        )

        mock_requests.add_callback(
            responses.GET,
            re.compile(self.url + "/files/.*"),
            callback=self.download_file_callback
        )

        mock_requests.assert_all_requests_are_fired = False

    def check_auth_header(self, request):
        auth_hdr = request.headers['Authorization']
        return auth_hdr is not None and auth_hdr == 'Bearer ' + self.token

    def list_datasets_callback(self, request):
        if not self.check_auth_header(request):
            return 400, {}, json.dumps({"error_description": "invalid token"})

        return 200, {}, json.dumps(self.all_datasets)

    def list_files_callback(self, request):
        if not self.check_auth_header(request):
            return 400, {}, json.dumps({"error_description": "invalid token"})

        dataset_id = request.path_url.split('/')[-2]
        files = None
        if self.dataset_files[dataset_id] is not None:
            files = [self.files[file_id] for file_id in sorted(self.dataset_files[dataset_id])]

        return 200, {}, json.dumps(files)

    def get_file_metadata_callback(self, request):
        if not self.check_auth_header(request):
            return 400, {}, json.dumps({"error_description": "invalid token"})

        try:
            file_id = urllib.parse.urlsplit(request.url).path.split('/')[-1]
            return 200, {}, json.dumps(self.files[file_id])
        except KeyError:
            return 404, {}, None

    @staticmethod
    def parse_ranges(s):
        return tuple(map(int, re.match(r'^bytes=(\d+)-(\d+)$', s).groups()))

    def download_file_callback(self, request):
        if not self.check_auth_header(request):
            return 400, {}, json.dumps({"error_description": "invalid token"})

        file_id = urllib.parse.urlsplit(request.url).path.split('/')[-1]
        file_content = self.file_content[file_id]

        if request.headers['Range'] is not None:
            start, end = self.parse_ranges(request.headers['Range'])
            assert start < end
            return 200, {}, file_content[start:end + 1]

        return 200, {}, file_content

    @property
    def all_datasets(self):
        if self.dataset_files is None:
            return None
        return list(sorted(self.dataset_files.keys()))



import os
import random
from unittest import mock

from psutil import virtual_memory

from pyega3.libs import utils


@mock.patch('os.remove')
def test_merge_bin_files_on_disk(mocked_remove):
    mem = virtual_memory().available
    files_to_merge = {
        'f1.bin': os.urandom(random.randint(1, mem // 512)),
        'f2.bin': os.urandom(random.randint(1, mem // 512)),
        'f3.bin': os.urandom(random.randint(1, mem // 512)),
    }
    target_file_name = "merged.file"

    merged_bytes = bytearray()

    def mock_write(buf):
        merged_bytes.extend(buf)

    real_open = open

    def open_wrapper(filename, mode):
        if filename == target_file_name:
            file_object = mock.mock_open().return_value
            file_object.write.side_effect = mock_write
            return file_object
        if filename not in files_to_merge:
            return real_open(filename, mode)
        content = files_to_merge[filename]
        length = len(content)
        buf_size = 65536
        file_object = mock.mock_open(read_data=content).return_value
        file_object.__iter__.return_value = [content[i:min(i + buf_size, length)] for i in
                                             range(0, length, buf_size)]
        return file_object

    with mock.patch('builtins.open', new=open_wrapper):
        with mock.patch('os.rename', lambda s, d: merged_bytes.extend(files_to_merge[os.path.basename(s)])):
            utils.merge_bin_files_on_disk(target_file_name, list(files_to_merge.keys()),
                                          0)  # this value can be changed from 0 to other/actual value

    mocked_remove.assert_has_calls([mock.call(f) for f in list(files_to_merge.keys())[1:]])

    verified_bytes = 0
    for f_content in files_to_merge.values():
        f_len = len(f_content)
        assert f_content == merged_bytes[verified_bytes: verified_bytes + f_len]
        verified_bytes += f_len

    assert verified_bytes == len(merged_bytes)



import hashlib
import math
import os
import tempfile
from collections import namedtuple
from unittest import mock

import pytest

from pyega3.libs.data_file import DataFile

OUTPUT_DIR = tempfile.gettempdir()


@pytest.fixture
def mock_writing_files():
    files = {}

    def open_wrapper(filename, mode):
        filename = os.path.basename(filename)
        if filename not in files:
            if 'r' in mode:
                raise Exception("Attempt to read mock file before it was created.")
            files[filename] = bytearray()
        content = bytes(files[filename])
        content_len = len(content)
        read_buf_sz = 65536
        file_object = mock.mock_open(read_data=content).return_value
        file_object.__iter__.return_value = [content[i:min(i + read_buf_sz, content_len)] for i in
                                             range(0, content_len, read_buf_sz)]
        file_object.write.side_effect = lambda write_buf: files[filename].extend(write_buf)
        return file_object

    def os_stat_mock(fn):
        fn = os.path.basename(fn)
        X = namedtuple('X', 'st_size f1 f2 f3 f4 f5 f6 f7 f8 f9')
        result = X(*([None] * 10))
        return result._replace(st_size=len(files[fn]))

    def os_rename_mock(s, d):
        files.__setitem__(os.path.basename(d), files.pop(os.path.basename(s)))

    with mock.patch('builtins.open', new=open_wrapper):
        with mock.patch('os.makedirs', return_value=None):
            with mock.patch('os.path.exists', lambda path: os.path.basename(path) in files):
                with mock.patch('os.stat', os_stat_mock):
                    with mock.patch('os.rename', os_rename_mock):
                        with mock.patch('shutil.rmtree'):
                            with mock.patch('os.listdir', return_value=[]):
                                yield files


def test_download_file(mock_data_server, random_binary_file, mock_writing_files, mock_server_config, mock_data_client):
    file_id = "EGAF00000000001"
    file_name = "resulting.file"
    file_md5 = hashlib.md5(random_binary_file).hexdigest()

    mock_data_server.file_content[file_id] = random_binary_file

    file = DataFile(mock_data_client, file_id, display_file_name=file_name, file_name=file_name + ".cip",
                    size=len(random_binary_file) + 16, unencrypted_checksum=file_md5)
    file.download_file_retry(1, output_dir=OUTPUT_DIR, genomic_range_args=None, max_retries=5, retry_wait=0)
    assert random_binary_file == mock_writing_files[file_name]


def test_no_error_if_output_file_already_exists_with_correct_md5(mock_data_server, random_binary_file,
                                                                 mock_writing_files, mock_server_config,
                                                                 mock_data_client):
    file_id = "EGAF00000000001"
    file_name = "resulting.file"
    file_md5 = hashlib.md5(random_binary_file).hexdigest()

    mock_data_server.file_content[file_id] = random_binary_file

    mock_writing_files[file_name] = random_binary_file

    # add 16 bytes to file size ( IV adjustment )
    file = DataFile(mock_data_client, file_id, display_file_name=file_name, file_name=file_name + ".cip",
                    size=len(random_binary_file) + 16, unencrypted_checksum=file_md5)
    file.download_file_retry(1,
                             output_dir=OUTPUT_DIR,
                             genomic_range_args=None, max_retries=5, retry_wait=0)


def test_output_file_is_removed_if_md5_was_invalid(mock_data_server, random_binary_file, mock_writing_files,
                                                   mock_server_config,
                                                   mock_data_client):
    file_id = "EGAF00000000001"
    file_name = "resulting.file"
    wrong_md5 = "wrong_md5_exactly_32_chars_longg"

    mock_data_server.file_content[file_id] = random_binary_file

    file = DataFile(mock_data_client, file_id, file_name, file_name + ".cip", len(random_binary_file) + 16, wrong_md5)

    with mock.patch('os.remove') as mocked_remove:
        with pytest.raises(Exception):
            file.download_file_retry(1, OUTPUT_DIR, genomic_range_args=None, max_retries=5, retry_wait=0)

    mocked_remove.assert_has_calls(
        [mock.call(os.path.join(os.getcwd(), file_id, os.path.basename(f))) for f in
         list(mock_writing_files.keys()) if file_name not in f],
        any_order=True)


def test_genomic_range_calls_htsget(mock_data_server, random_binary_file, mock_writing_files, mock_server_config,
                                    mock_data_client):
    file_id = "EGAF00000000001"
    file_name = "resulting.file"
    file_md5 = hashlib.md5(random_binary_file).hexdigest()

    mock_data_server.file_content[file_id] = random_binary_file

    file = DataFile(mock_data_client, file_id, file_name, file_name + ".cip", len(random_binary_file) + 16, file_md5)

    with mock.patch('htsget.get') as mocked_htsget:
        file.download_file_retry(
            1, output_dir=OUTPUT_DIR, genomic_range_args=("chr1", None, 1, 100, None),
            max_retries=5,
            retry_wait=0)

    args, kwargs = mocked_htsget.call_args
    assert args[0] == f'{mock_server_config.url_api_ticket}/files/EGAF00000000001'

    assert kwargs.get('reference_name') == 'chr1'
    assert kwargs.get('reference_md5') is None
    assert kwargs.get('start') == 1
    assert kwargs.get('end') == 100
    assert kwargs.get('data_format') is None


def test_gpg_files_not_supported(mock_data_client):
    file = DataFile(mock_data_client, "", "test.gz", "test.gz.gpg", 0, "")

    file.download_file_retry(1, output_dir=OUTPUT_DIR, genomic_range_args=None, max_retries=5, retry_wait=5)


def test_temporary_chunk_files_stored_in_temp_folder_with_suffix_tmp(mock_data_server, random_binary_file,
                                                                     mock_server_config,
                                                                     mock_data_client):
    # Given: a file that exist in EGA object store and the user has permissions to access to it
    file_id = "EGAF00000000001"
    file_name = "resulting.file"
    file_md5 = hashlib.md5(random_binary_file).hexdigest()

    mock_data_server.file_content[file_id] = random_binary_file

    file = DataFile(mock_data_client, file_id, file_name, file_name + ".cip", len(random_binary_file) + 16, file_md5)

    # When: the user starts to download a file
    output_file = os.path.join(OUTPUT_DIR, file_id, file_name)
    md5_file = output_file + ".md5"
    if os.path.exists(output_file):
        os.remove(output_file)
    if os.path.exists(md5_file):
        os.remove(md5_file)

    with mock.patch('builtins.open', wraps=open) as wrapped_open:
        file.download_file_retry(1, output_dir=OUTPUT_DIR, genomic_range_args=None, max_retries=5, retry_wait=0)

    # Then: The temporary files for the chunks are in the temporary folder and has .tmp as a suffix
    temporary_folder = os.path.join(OUTPUT_DIR, file_id, ".tmp_download")
    # call[1] is a list of positional arguments which were passed to open():
    slices_opened = set([call[1][0] for call in wrapped_open.mock_calls if len(call[1]) == 2])
    slices_opened.remove(output_file)
    slices_opened.remove(md5_file)

    for slice_file in slices_opened:
        assert slice_file.startswith(temporary_folder)
        assert slice_file.endswith(".tmp")


# Feature: The user can configure the slice sizes used when downloading a file.

def test_the_user_specifies_a_slice_size(mock_data_client):
    # Given: a file that the user has permissions to download and a custom slice size
    file = DataFile(mock_data_client, file_id="EGAF123456", size=12345, unencrypted_checksum="testChecksum")
    slice_size = 1000

    # When: when the user downloads the file
    with mock.patch("pyega3.libs.data_file.DataFile.download_file_slice") as mock_download_slice:
        with mock.patch("pyega3.libs.utils.md5", return_value=file.unencrypted_checksum):
            with mock.patch("os.path.getsize", return_value=file.size):
                file.download_file(output_file="output_file", num_connections=1, max_slice_size=slice_size)

    # Then: the file is downloaded in multiple slices where each slice is at most the custom slice size
    assert mock_download_slice.call_count == 13


def test_the_user_does_not_specifies_a_slice_size(mock_data_client):
    # Given: a file that the user has permissions to download
    file = DataFile(mock_data_client, file_id="EGAF123456", size=1234567890, unencrypted_checksum="testChecksum")

    # When: when the user downloads the file
    with mock.patch("pyega3.libs.data_file.DataFile.download_file_slice") as mock_download_slice:
        with mock.patch("pyega3.libs.utils.md5", return_value=file.unencrypted_checksum):
            with mock.patch("os.path.getsize", return_value=file.size):
                file.download_file(output_file="output_file", num_connections=1)

    # Then: The file is downloaded in multiple slices where each slice is at most the default slice size
    assert mock_download_slice.call_count == math.ceil(file.size / DataFile.DEFAULT_SLICE_SIZE)


def test_the_user_specifies_a_custom_slice_size_different_to_before(mock_data_client, mock_data_server,
                                                                    random_binary_file, caplog):
    # Given: a file that the user has permissions to download and a custom slice size and some slices that were already downloaded with different size.
    mock_data_server.file_content["EGAF123456"] = random_binary_file
    file = DataFile(mock_data_client, file_id="EGAF123456", size=12345, unencrypted_checksum="testChecksum")
    slice_size = 1000
    os.makedirs(".tmp_download", exist_ok=True)

    extra_slice = file.download_file_slice(f'.tmp_download/{file.id}', 0, 1234)
    assert os.path.exists(extra_slice)

    # When: when the user downloads the file
    with mock.patch("pyega3.libs.data_file.DataFile.download_file_slice") as mock_download_slice:
        with mock.patch("pyega3.libs.utils.md5", return_value=file.unencrypted_checksum):
            with mock.patch("os.path.getsize", return_value=file.size):
                file.download_file(output_file="output_file", num_connections=1, max_slice_size=slice_size)

    # Then: the file is downloaded in multiple slices where each slice is at most the custom slice size and delete the old slices with the warning.
    assert mock_download_slice.call_count == 13
    assert not os.path.exists(extra_slice)
    assert "Deleting the leftover" in caplog.text


def test_slice_file_is_reused(mock_data_client, mock_data_server, random_binary_file, caplog):
    # Given: a file that the user has permissions to
    # download and a custom slice size and some slices that
    # were already downloaded with correct size.
    mock_data_server.file_content["EGAF123456"] = random_binary_file
    file = DataFile(mock_data_client, file_id="EGAF123456", size=12345, unencrypted_checksum="testChecksum")
    slice_size = 1000
    os.makedirs(".tmp_download", exist_ok=True)

    extra_slice = file.download_file_slice(f'.tmp_download/{file.id}', 0, slice_size)
    assert os.path.exists(extra_slice)

    # When: when the user downloads the file
    with mock.patch("pyega3.libs.data_file.DataFile.download_file_slice") as mock_download_slice:
        with mock.patch("pyega3.libs.utils.md5", return_value=file.unencrypted_checksum):
            with mock.patch("os.path.getsize", return_value=file.size):
                file.download_file(output_file="output_file", num_connections=1, max_slice_size=slice_size)

    # Then: the file is downloaded in multiple
    # slices where each slice is at most the custom
    # slice size and old slices is not deleted.
    assert mock_download_slice.call_count == 13
    assert os.path.exists(extra_slice)
    assert "Deleting the leftover" not in caplog.text


def teardown_module():
    filepath = 'output_file.md5'
    if os.path.exists(filepath):
        os.remove(filepath)



import json
from unittest import mock

import pytest

from pyega3.libs.credentials import Credentials
import tests.unit.conftest as common


def test_load_username_and_password_from_credentials_file(mock_input_file):
    username = common.rand_str()
    password = common.rand_str()

    with mock_input_file(json.dumps({"username": username, "password": password})) as credentials_file:
        result = Credentials.from_file(credentials_file)
        assert result.username == username
        assert result.password == password


def test_when_credentials_file_has_no_password_ask_user_for_it(mock_input_file):
    username = common.rand_str()
    password = common.rand_str()

    with mock_input_file(json.dumps({"username": username})) as credentials_file, mock.patch("getpass.getpass",
                                                                                             return_value=password):
        result = Credentials.from_file(credentials_file)
        assert result.username == username
        assert result.password == password


def test_error_when_credentials_file_is_bad_json(mock_input_file):
    with mock_input_file("bad json") as credentials_file:
        with pytest.raises(SystemExit):
            Credentials.from_file(credentials_file)


def test_get_credential_prompts_user_for_username_and_password():
    username = common.rand_str()
    password = common.rand_str()

    with mock.patch('builtins.input', return_value=username):
        with mock.patch('getpass.getpass', return_value=password):
            credentials = Credentials()
            credentials.prompt_for_missing_values()
            assert credentials.username == username
            assert credentials.password == password


def test_load_credential_prompts_user_for_credentials_if_credentials_file_does_not_exist():
    username = common.rand_str()
    password = common.rand_str()

    with mock.patch('builtins.input', return_value=username):
        with mock.patch('getpass.getpass', return_value=password):
            with mock.patch('os.path.exists', return_value=False):
                credentials = Credentials.from_file("unknownfile.txt")
                assert credentials.username == username
                assert credentials.password == password



import logging
import os

import pytest

from pyega3.libs import utils
from pyega3.libs.data_set import DataSet

OUTPUT_DIR = 'user_specified_directory'


def file_and_md5_file_paths(mock_dataset_file):
    return [
        f'/{OUTPUT_DIR}/{mock_dataset_file["fileId"]}/{mock_dataset_file["displayFileName"]}',
        f'/{OUTPUT_DIR}/{mock_dataset_file["fileId"]}/{mock_dataset_file["displayFileName"]}.md5'
    ]


@pytest.fixture()
def expected_file_paths(mock_dataset_files):
    nested_paths = [file_and_md5_file_paths(file) for file in mock_dataset_files]

    def flatten(nested_list):
        return [element for sublist in nested_list for element in sublist]

    return flatten(nested_paths)


@pytest.fixture()
def dataset_in_fire(mock_data_client, dataset_with_files):
    return DataSet(mock_data_client, dataset_with_files.id)


def download_dataset_from_fire(dataset_in_fire):
    # num_connections, output_dir, genomic_range_args, max_retries=5, retry_wait=5
    dataset_in_fire.download(1, OUTPUT_DIR, None, max_retries=2, retry_wait=0)


def assert_that_expected_file_paths_not_exist(expected_file_paths):
    for path in expected_file_paths:
        assert not os.path.exists(path)


def assert_that_expected_file_paths_exist(expected_file_paths):
    for path in expected_file_paths:
        assert os.path.isfile(path)


def test_file_is_saved_into_an_existing_directory_which_was_specified_by_the_user(dataset_in_fire,
                                                                                  expected_file_paths,
                                                                                  fs):
    """
    The user wants to download a file into a specific, user-specified directory.
    The directory exists and there is no such file in that directory yet.
    Expected: the file is downloaded into the user-specified directory
    and the file-name will be the 'displayFileName' (e.g. 'EGAF00000000001/ENCFF284YOU.bam.bai').
    """

    os.makedirs(OUTPUT_DIR)

    assert os.path.isdir(OUTPUT_DIR)
    assert_that_expected_file_paths_not_exist(expected_file_paths)

    download_dataset_from_fire(dataset_in_fire)

    assert_that_expected_file_paths_exist(expected_file_paths)


def test_file_in_directory_is_not_downloaded_again(mock_dataset_files, dataset_in_fire, fs, caplog):
    """
    When the file has correctly been downloaded already into the specified directory,
    then it is not downloaded again.
    """
    caplog.set_level(logging.INFO)
    mtimes = {}

    # First I create the downloaded files to simulate a previous, successful download:
    for file in mock_dataset_files:
        file_path, md5_file_path = file_and_md5_file_paths(file)
        fs.create_file(file_path)

        with open(file_path, 'wb') as fh:
            fh.write(file['fileContent'])
        with open(md5_file_path, 'w') as fh:
            fh.write(file['unencryptedChecksum'])

        # And I store their original mtimes:
        mtimes[file_path] = os.stat(file_path).st_mtime
        mtimes[md5_file_path] = os.stat(md5_file_path).st_mtime

        # The downloaded files should exist:
        assert os.path.isfile(file_path)
        assert os.path.isfile(md5_file_path)

    assert os.path.isdir(OUTPUT_DIR)

    # Let's simulate a second download into the specified directory.
    # This should not download again the files,
    # but it should print out a "Local file exists" message.
    download_dataset_from_fire(dataset_in_fire)

    for file_path, orig_mtime_before_second_download in mtimes.items():
        mtime_after_second_download = os.stat(file_path).st_mtime
        # If the files haven't been re-downloaded again, then the mtimes should not have changed:
        assert mtime_after_second_download == orig_mtime_before_second_download

        if not file_path.endswith('.md5'):
            assert f"Local file exists:'{file_path}'" in caplog.text


def test_corrupted_file_in_a_directory_is_downloaded_again(mock_dataset_files, dataset_in_fire, fs):
    """
    When the file has been downloaded incorrectly into a directory previously, then it is re-downloaded again.
    I don't think this happens too often, because an incorrectly downloaded file is always removed
    (or at least, it should be removed always).
    """
    # First I create the downloaded files to simulate a previous, successful download:
    for file in mock_dataset_files:
        file_path, md5_file_path = file_and_md5_file_paths(file)

        # First I create a file to simulate a previous, unsuccessful download:
        fs.create_file(file_path)
        with open(file_path, 'w') as fh:
            fh.write('incorrect, corrupted content')

        # The downloaded file should exist, but without an .md5:
        assert os.path.isfile(file_path)
        assert not os.path.exists(md5_file_path)

    assert os.path.isdir(OUTPUT_DIR)

    # Let's simulate a second download into the specified directory.
    # This should download again the file and the .md5, this time, correctly.
    download_dataset_from_fire(dataset_in_fire)

    for file in mock_dataset_files:
        file_path, md5_file_path = file_and_md5_file_paths(file)

        # Check that the actual MD5 of the re-downloaded file is correct:
        correct_expected_md5 = file['unencryptedChecksum']
        actual_md5 = utils.calculate_md5(file_path, file['fileSize'])
        assert actual_md5 == correct_expected_md5

        # The .md5 file should now exist:
        assert os.path.isfile(md5_file_path)

        # And the content of the .md5 file should be the correct, expected one:
        with open(md5_file_path, 'r') as fh:
            actual_md5 = fh.readline()
        assert actual_md5 == correct_expected_md5


def test_if_specified_dir_is_a_file_then_error_is_thrown(dataset_in_fire, expected_file_paths, caplog, fs):
    """
    The user wants to download a file into a specific, user-specified directory,
    but that directory is, in fact, a file. An error is thrown and nothing is downloaded.
    """
    caplog.set_level(logging.INFO)

    existing_file = OUTPUT_DIR
    fs.create_file(existing_file)
    assert os.path.exists(existing_file)
    assert os.path.isfile(existing_file)

    assert_that_expected_file_paths_not_exist(expected_file_paths)

    download_dataset_from_fire(dataset_in_fire)

    assert f"Not a directory in the fake filesystem: '/{OUTPUT_DIR}'" in caplog.text

    assert_that_expected_file_paths_not_exist(expected_file_paths)


def test_non_existing_directory_is_created(dataset_in_fire, expected_file_paths, fs):
    """
    The user wants to download a file into a specific, user-specified directory.
    The directory does not exist.
    Expected: the non-existent directory is created automatically for the user,
    the file is downloaded into the user-specified directory
    and the file-name will be the 'displayFileName' (e.g. 'EGAF00000000001/ENCFF284YOU.bam.bai').
    The non-existent directory is created automatically as a (positive) side-effect of the code
    which creates the .tmp_download directory: that code creates not only the .tmp_download directory,
    but all the missing parent directories, too, including the non-existent user-specified directory.
    """

    assert not os.path.exists(OUTPUT_DIR)
    assert_that_expected_file_paths_not_exist(expected_file_paths)

    download_dataset_from_fire(dataset_in_fire)

    assert_that_expected_file_paths_exist(expected_file_paths)



import subprocess


def run(command: str, input: str = None, verbose: bool = True):
    parsed_command = command.split(' ')
    proc = subprocess.Popen(parsed_command,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            )
    if input:
        stdout, stderr = proc.communicate(input.encode())
    else:
        stdout, stderr = proc.communicate()

    if verbose:
        print('$ ' + command)
        print(stdout.decode())
        print(stderr.decode())

    return proc.returncode, stdout.decode(), stderr.decode()



import glob
import random
from collections import namedtuple
from unittest import mock
from unittest.mock import Mock, patch

import pytest
import requests
import os

from urllib3.exceptions import NewConnectionError

import tests.unit.conftest as common
from pyega3.libs.data_file import DataFile


@pytest.fixture
def slice_file(random_binary_file):
    file_id = "EGAF1234"
    file_size = len(random_binary_file)
    slice_start = random.randint(0, file_size)
    slice_len = random.randint(0, file_size - slice_start)
    file_name = common.rand_str()
    file_name_for_slice = f'{file_name}-from-{slice_start}-len-{slice_len}.slice'
    SliceFile = namedtuple("SliceFile", "id original_file_size start length original_file_name file_name binary")
    return SliceFile(file_id, file_size, slice_start, slice_len, file_name, file_name_for_slice, random_binary_file)


def test_download_file_slice_downloads_correct_bytes_to_file(mock_data_server, slice_file, mock_data_client):
    mock_data_server.file_content[slice_file.id] = slice_file.binary
    written_bytes = 0

    def mock_write(buf):
        nonlocal written_bytes
        buf_len = len(buf)
        expected_buf = slice_file.binary[slice_file.start + written_bytes:slice_file.start + written_bytes + buf_len]
        assert expected_buf == buf
        written_bytes += buf_len

    file_name = common.rand_str()
    file_name_for_slice = file_name + '-from-' + str(slice_file.start) + '-len-' + str(
        slice_file.length) + '.slice.tmp'

    file = DataFile(mock_data_client, slice_file.id)

    m_open = mock.mock_open()
    with mock.patch("builtins.open", m_open, create=True):
        with mock.patch("os.path.getsize", lambda path: written_bytes if path == file_name_for_slice else 0):
            with mock.patch("os.rename"):
                m_open().write.side_effect = mock_write
                file.download_file_slice(file_name, slice_file.start, slice_file.length)
                assert slice_file.length == written_bytes

    m_open.assert_called_with(file_name_for_slice, 'ba')


def test_error_when_bad_token(mock_data_server, mock_data_client):
    file_id = "EGAF1234"
    bad_token = common.rand_str()
    mock_data_client.auth_client = common.MockAuthClient(bad_token)
    file = DataFile(mock_data_client, file_id)
    with pytest.raises(requests.exceptions.HTTPError):
        file.download_file_slice(common.rand_str(), 1, 10)


def test_error_when_bad_url(mock_data_client):
    file = DataFile(mock_data_client, "bad/url")
    with pytest.raises(NewConnectionError):
        file.download_file_slice(common.rand_str(), 1, 10)


def test_error_when_start_is_negative(mock_data_client):
    file = DataFile(mock_data_client, common.rand_str())
    with pytest.raises(ValueError):
        file.download_file_slice(common.rand_str(), -1, 1)


def test_error_when_end_is_negative(mock_data_client):
    file = DataFile(mock_data_client, common.rand_str())
    with pytest.raises(ValueError):
        file.download_file_slice(common.rand_str(), 0, -1)


def test_slice_file_name_removes_tmp_suffix_when_successful(mock_data_server, mock_data_client, slice_file):
    # Given: a file that exist in EGA object store and the user has permissions to access to it
    mock_data_server.file_content[slice_file.id] = slice_file.binary

    # When: the user successfully downloads a chunk
    file = DataFile(mock_data_client, slice_file.id)
    file.download_file_slice(slice_file.original_file_name, slice_file.start, slice_file.length)

    # Then: the suffix .tmp is removed from file for the successful chunk
    assert os.path.exists(slice_file.file_name)
    assert not os.path.exists(slice_file.file_name + '.tmp')


def test_chunk_fails_to_download(mock_data_server, mock_data_client, slice_file):
    # Given: a file that exist in EGA object store and the user has permissions to access to it
    mock_data_server.file_content[slice_file.id] = slice_file.binary

    slice_length = len(slice_file.binary) + 10

    # When: the user unsuccessfully downloads a chunk
    file_name = common.rand_str()
    file = DataFile(mock_data_client, slice_file.id)
    try:
        file.download_file_slice(file_name, slice_file.start, slice_length)
    except:
        # For the purpose of this test the download should fail
        pass

    # Then: file for the failed chunk is removed
    file_name_for_slice = file_name + '-from-' + str(slice_file.start) + '-len-' + str(slice_length) + '.slice'
    assert not os.path.exists(file_name_for_slice)
    assert not os.path.exists(file_name_for_slice + '.tmp')


def test_return_slice_file_when_existing(mock_data_server, mock_data_client, slice_file):
    # Given: a slice file existing in tmp directory
    mock_data_server.file_content[slice_file.id] = slice_file.binary
    file = DataFile(mock_data_client, slice_file.id)

    mock_stat = Mock()
    mock_stat.st_size = slice_file.length

    with patch.object(mock_data_client, 'get_stream', wraps=mock_data_client.get_stream) as get_stream_mock, \
        mock.patch("os.path.exists", lambda path: True if path == slice_file.file_name else False), \
        mock.patch("os.path.getsize", lambda path: slice_file.length), \
        mock.patch("os.stat", lambda path: mock_stat):
        # When: the slice file is downloaded
        filename = file.download_file_slice(slice_file.original_file_name, slice_file.start, slice_file.length)
        # Then: the existing slice file with same length is reused and data is not re-fetched
        assert filename == slice_file.file_name
        get_stream_mock.assert_not_called()


def test_remove_existing_slice_file_when_it_exceeds_slice_length(mock_data_server, mock_data_client, slice_file):
    # Given: a slice file existing in tmp directory whose size exceeds the expected slice length
    mock_data_server.file_content[slice_file.id] = slice_file.binary
    file = DataFile(mock_data_client, slice_file.id)

    mock_stat = Mock()
    mock_stat.st_size = slice_file.length + 1

    with mock.patch("os.remove") as remove_file_mock, \
        mock.patch("os.path.exists", lambda path: True if path == slice_file.file_name else False), \
        mock.patch("os.path.getsize", lambda path: slice_file.length), \
        mock.patch("os.stat", lambda path: mock_stat):
        # When: the slice file is downloaded
        output_file = file.download_file_slice(slice_file.original_file_name, slice_file.start, slice_file.length)
        # Then: the existing slice file is deleted
        assert output_file == slice_file.file_name
        remove_file_mock.assert_called_once_with(slice_file.file_name)


def teardown_module():
    for f in glob.glob(f'{os.getcwd()}/*.slice'):
        os.remove(f)



from collections import namedtuple

import pytest
import requests

import tests.unit.conftest as common
from pyega3.libs.data_file import DataFile


@pytest.fixture
def dummy_file(mock_data_server):
    file_id = "EGAF00000000001"

    file_size = 4804928
    file_name = "EGAZ00000000001/ENCFF000001.bam"
    display_file_name = "ENCFF000001.bam"
    check_sum = "3b89b96387db5199fef6ba613f70e27c"

    file = {"fileId": file_id,
            "displayFileName": display_file_name,
            "fileName": file_name,
            "fileSize": file_size,
            "unencryptedChecksum": check_sum,
            "fileStatus": "available"}

    mock_data_server.files[file_id] = file

    DummyFile = namedtuple("DummyFile", ["id", "file_size", "file_name", "display_file_name", "check_sum"])
    return DummyFile(file_id, file_size, file_name, display_file_name, check_sum)


def test_get_file_name_size_md5(mock_data_server, dummy_file, mock_data_client):
    file = DataFile(mock_data_client, dummy_file.id)
    assert file.display_name == dummy_file.display_file_name
    assert file.name == dummy_file.file_name
    assert file.size == dummy_file.file_size
    assert file.unencrypted_checksum == dummy_file.check_sum


def test_error_with_bad_token(mock_data_server, dummy_file, mock_data_client):
    bad_token = common.rand_str()
    mock_data_client.auth_client = common.MockAuthClient(bad_token)
    with pytest.raises(requests.exceptions.HTTPError):
        file = DataFile(mock_data_client, dummy_file.id)
        file.load_metadata()


def test_error_with_unknown_file(mock_data_server, mock_data_client):
    bad_file_id = "EGAF00000000000"
    with pytest.raises(requests.exceptions.HTTPError):
        file = DataFile(mock_data_client, bad_file_id)
        file.load_metadata()


def test_error_with_file_with_bad_metadata(mock_data_server, mock_data_client):
    bad_file_id = "EGAF00000000666"
    mock_data_server.files[bad_file_id] = {"fileName": None, "displayFileName": None, "unencryptedChecksum": None}
    with pytest.raises(RuntimeError):
        file = DataFile(mock_data_client, bad_file_id)
        file.load_metadata()



import responses

from pyega3.libs.utils import get_client_ip


def test_when_ipinfo_is_blocked_return_unknown(mock_requests):
    endpoint = 'https://ipinfo.io/json'
    mock_requests.add(responses.GET, endpoint, status=403)

    resp_ip = get_client_ip()

    assert resp_ip == 'Unknown'



import concurrent.futures
import logging
import logging.handlers
import os
import re
import shutil
import sys
import time
import urllib

import htsget
import psutil
from tqdm import tqdm

from pyega3.libs import utils

DOWNLOAD_FILE_MEMORY_BUFFER_SIZE = 32 * 1024


class DataFile:
    DEFAULT_SLICE_SIZE = 100 * 1024 * 1024
    temporary_files_should_be_deleted = False

    def __init__(self, data_client, file_id,
                 display_file_name=None,
                 file_name=None,
                 size=None,
                 unencrypted_checksum=None,
                 status=None):
        self.data_client = data_client
        self.id = file_id

        self.temporary_files = set()

        self._display_file_name = display_file_name
        self._file_name = file_name
        self._file_size = size
        self._unencrypted_checksum = unencrypted_checksum
        self._file_status = status

    def load_metadata(self):
        res = self.data_client.get_json(f"/metadata/files/{self.id}")

        # If the user does not have access to the file then the server returns HTTP code 200 but the JSON payload has
        # all the fields empty
        if res['displayFileName'] is None or res['unencryptedChecksum'] is None:
            raise RuntimeError(f"Metadata for file id '{self.id}' could not be retrieved. " +
                               "This is probably because your account does not have access to this file. "
                               "You can check which datasets your account has access to at "
                               "'https://ega-archive.org/my-datasets.php' after logging in.")

        self._display_file_name = res['displayFileName']
        self._file_name = res['fileName']
        self._file_size = res['fileSize']
        self._unencrypted_checksum = res['unencryptedChecksum']
        self._file_status = res['fileStatus']

    @property
    def display_name(self):
        if self._display_file_name is None:
            self.load_metadata()
        return self._display_file_name

    @property
    def name(self):
        if self._file_name is None:
            self.load_metadata()
        return self._file_name

    @property
    def size(self):
        if self._file_size is None:
            self.load_metadata()
        return self._file_size

    @property
    def unencrypted_checksum(self):
        if self._unencrypted_checksum is None:
            self.load_metadata()
        return self._unencrypted_checksum

    @property
    def status(self):
        if self._file_status is None:
            self.load_metadata()
        return self._file_status

    @staticmethod
    def print_local_file_info(prefix_str, file, md5):
        logging.info(f"{prefix_str}'{os.path.abspath(file)}'({os.path.getsize(file)} bytes, md5={md5})")

    def download_file(self, output_file, num_connections=1, max_slice_size=DEFAULT_SLICE_SIZE):
        """Download an individual file"""

        file_size = self.size
        check_sum = self.unencrypted_checksum
        options = {"destinationFormat": "plain"}

        file_size -= 16  # 16 bytes IV not necesary in plain mode

        if os.path.exists(output_file) and utils.md5(output_file, file_size) == check_sum:
            DataFile.print_local_file_info('Local file exists:', output_file, check_sum)
            return

        num_connections = max(num_connections, 1)
        num_connections = min(num_connections, 128)

        if file_size < 100 * 1024 * 1024:
            num_connections = 1

        logging.info(f"Download starting [using {num_connections} connection(s), file size {file_size} and chunk "
                     f"length {max_slice_size}]...")

        chunk_len = max_slice_size

        temporary_directory = os.path.join(os.path.dirname(output_file), ".tmp_download")
        os.makedirs(temporary_directory, exist_ok=True)

        with tqdm(total=int(file_size), unit='B', unit_scale=True) as pbar:
            params = [
                (os.path.join(temporary_directory, self.id), chunk_start_pos,
                 min(chunk_len, file_size - chunk_start_pos), options, pbar)
                for chunk_start_pos in range(0, file_size, chunk_len)]

            for file in os.listdir(temporary_directory):
                match = re.match(r"(.*)-from-(\d*)-len-(\d*).*", file)
                file_id = match.group(1)
                file_from = match.group(2)
                file_length = match.group(3)

                if file_id != self.id:
                    continue

                if (int(file_from), int(file_length)) in [(param[1], param[2]) for param in params]:
                    continue

                logging.warning(f'Deleting the leftover {file} temporary file because the MAX_SLICE_SIZE parameter ('
                                f'and thus the slice sizes) have been modified since the last run.')
                os.remove(os.path.join(temporary_directory, file))

            results = []
            with concurrent.futures.ThreadPoolExecutor(max_workers=num_connections) as executor:
                for part_file_name in executor.map(self.download_file_slice_, params):
                    results.append(part_file_name)

            pbar.close()

            downloaded_file_total_size = sum(os.path.getsize(f) for f in results)
            if downloaded_file_total_size == file_size:
                utils.merge_bin_files_on_disk(output_file, results, downloaded_file_total_size)

        not_valid_server_md5 = len(str(check_sum or '')) != 32

        logging.info("Calculating md5 (this operation can take a long time depending on the file size)")

        received_file_md5 = utils.md5(output_file, file_size)

        logging.info("Verifying file checksum")

        if received_file_md5 == check_sum or not_valid_server_md5:
            DataFile.print_local_file_info('Saved to : ', output_file, check_sum)
            if not_valid_server_md5:
                logging.info(
                    f"WARNING: Unable to obtain valid MD5 from the server (received: {check_sum})."
                    f" Can't validate download. Please contact EGA helpdesk on helpdesk@ega-archive.org")
            with open(utils.get_fname_md5(output_file), 'wb') as f:  # save good md5 in aux file for future re-use
                f.write(received_file_md5.encode())
        else:
            os.remove(output_file)
            raise Exception(f"Download process expected md5 value '{check_sum}' but got '{received_file_md5}'")

    def download_file_slice_(self, args):
        return self.download_file_slice(*args)

    def download_file_slice(self, file_name, start_pos, length, options=None, pbar=None):
        if start_pos < 0:
            raise ValueError("start : must be positive")
        if length <= 0:
            raise ValueError("length : must be positive")

        path = f"/files/{self.id}"
        if options is not None:
            path += '?' + urllib.parse.urlencode(options)

        final_file_name = f'{file_name}-from-{str(start_pos)}-len-{str(length)}.slice'
        file_name = final_file_name + '.tmp'

        self.temporary_files.add(file_name)

        existing_size = os.stat(final_file_name).st_size if os.path.exists(final_file_name) else 0
        if existing_size > length:
            os.remove(final_file_name)
            existing_size = 0

        if pbar:
            pbar.update(existing_size)

        if existing_size == length:
            return final_file_name

        try:
            range_start = start_pos + existing_size
            range_end = start_pos + length - 1
            extra_headers = {
                'Range': f'bytes={range_start}-{range_end}'
            }
            with self.data_client.get_stream(path, extra_headers) as r:
                with open(file_name, 'ba') as file_out:
                    for chunk in r.iter_content(DOWNLOAD_FILE_MEMORY_BUFFER_SIZE):
                        file_out.write(chunk)
                        if pbar:
                            pbar.update(len(chunk))

            total_received = os.path.getsize(file_name)
            if total_received != length:
                raise Exception(f"Slice error: received={total_received}, requested={length}, file='{file_name}'")

        except Exception:
            if os.path.exists(file_name):
                os.remove(file_name)
            raise

        os.rename(file_name, final_file_name)

        return final_file_name

    @staticmethod
    def is_genomic_range(genomic_range_args):
        if not genomic_range_args:
            return False
        return genomic_range_args[0] is not None or genomic_range_args[1] is not None

    def generate_output_filename(self, folder, genomic_range_args):
        file_name = self.display_name
        ext_to_remove = ".cip"
        if file_name.endswith(ext_to_remove):
            file_name = file_name[:-len(ext_to_remove)]
        name, ext = os.path.splitext(os.path.basename(file_name))

        genomic_range = ''
        if DataFile.is_genomic_range(genomic_range_args):
            genomic_range = "_genomic_range_" + (genomic_range_args[0] or genomic_range_args[1])
            genomic_range += '_' + (str(genomic_range_args[2]) or '0')
            genomic_range += '_' + (str(genomic_range_args[3]) or '')
            format_ext = '.' + (genomic_range_args[4] or '').strip().lower()
            if format_ext != ext and len(format_ext) > 1:
                ext += format_ext

        ret_val = os.path.join(folder, self.id, name + genomic_range + ext)
        logging.debug(f"Output file:'{ret_val}'")
        return ret_val

    @staticmethod
    def print_local_file_info_genomic_range(prefix_str, file, gr_args):
        logging.info(
            f"{prefix_str}'{os.path.abspath(file)}'({os.path.getsize(file)} bytes, referenceName={gr_args[0]},"
            f" referenceMD5={gr_args[1]}, start={gr_args[2]}, end={gr_args[3]}, format={gr_args[4]})"
        )

    def download_file_retry(self, num_connections, output_dir, genomic_range_args, max_retries, retry_wait,
                            max_slice_size=DEFAULT_SLICE_SIZE):
        if self.name.endswith(".gpg"):
            logging.info(
                "GPG files are currently not supported."
                " Please email EGA Helpdesk at helpdesk@ega-archive.org")
            return

        logging.info(f"File Id: '{self.id}'({self.size} bytes).")

        output_file = self.generate_output_filename(output_dir, genomic_range_args)

        temporary_directory = os.path.join(os.path.dirname(output_file), ".tmp_download")
        if not os.path.exists(temporary_directory):
            os.makedirs(temporary_directory)

        hdd = psutil.disk_usage(os.getcwd())
        logging.info(f"Total space : {hdd.total / (2 ** 30):.2f} GiB")
        logging.info(f"Used space : {hdd.used / (2 ** 30):.2f} GiB")
        logging.info(f"Free space : {hdd.free / (2 ** 30):.2f} GiB")

        # If file is bigger than free space, warning
        if hdd.free < self.size:
            logging.warning(f"The size of the file that you want to download is bigger than your free space in this "
                            f"location")

        if DataFile.is_genomic_range(genomic_range_args):
            with open(output_file, 'wb') as output:
                htsget.get(
                    f"{self.data_client.htsget_url}/files/{self.id}",
                    output,
                    reference_name=genomic_range_args[0], reference_md5=genomic_range_args[1],
                    start=genomic_range_args[2], end=genomic_range_args[3],
                    data_format=genomic_range_args[4],
                    max_retries=sys.maxsize if max_retries < 0 else max_retries,
                    retry_wait=retry_wait,
                    bearer_token=self.data_client.auth_client.token)
            DataFile.print_local_file_info_genomic_range('Saved to : ', output_file, genomic_range_args)
            return

        done = False
        num_retries = 0
        while not done:
            try:
                self.download_file(output_file, num_connections, max_slice_size)
                done = True
            except Exception as e:
                if e is ConnectionError:
                    logging.info("Failed to connect to data service. Check that the necessary ports are open in your "
                                 "firewall. See the documentation for more information.")
                logging.exception(e)
                if num_retries == max_retries:
                    if DataFile.temporary_files_should_be_deleted:
                        self.delete_temporary_folder(temporary_directory)

                    raise e
                time.sleep(retry_wait)
                num_retries += 1
                logging.info(f"retry attempt {num_retries}")

    def delete_temporary_folder(self, temporary_directory):
        try:
            shutil.rmtree(temporary_directory)
        except FileNotFoundError as ex:
            logging.error(f'Could not delete the temporary folder: {ex}')



import os

import pytest

from pyega3.libs import utils


@pytest.mark.parametrize("md5,data", [
    ("d41d8cd98f00b204e9800998ecf8427e", b""),
    ("0cc175b9c0f1b6a831c399e269772661", b"a"),
    ("900150983cd24fb0d6963f7d28e17f72", b"abc"),
    ("f96b697d7cb7938d525a2f31aaf161d0", b"message digest"),
    ("c3fcd3d76192e4007dfb496cca67e13b", b"abcdefghijklmnopqrstuvwxyz"),
    ("d174ab98d277d9f5a5611c2c9f419d9f", b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"),
    ("57edf4a22be3c955ac49da2e2107b67a",
     b"12345678901234567890123456789012345678901234567890123456789012345678901234567890")
])
def test_md5(md5, data, mock_input_file):
    with mock_input_file(data) as input_file:
        assert utils.md5(input_file, len(data)) == md5


def test_calculating_md5_of_non_existent_file_raises_exception():
    non_existent_file = '/tmp/non/existent/file'
    assert not os.path.exists(non_existent_file)

    with pytest.raises(Exception):
        utils.calculate_md5(non_existent_file, -1)



import getpass
import json
import logging
import os
import sys


class Credentials:
    def __init__(self, username=None, password=None):
        self.username = username
        self.password = password

    @staticmethod
    def from_file(filepath):
        """Load credentials for EMBL/EBI EGA from specified file"""
        result = Credentials()
        filepath = os.path.expanduser(filepath)
        if not os.path.exists(filepath):
            logging.error(f"{filepath} does not exist")
        else:
            try:
                with open(filepath) as f:
                    cfg = json.load(f)

                if 'username' in cfg:
                    result.username = cfg['username']
                if 'password' in cfg:
                    result.password = cfg['password']

            except ValueError:
                logging.error("Invalid credential config JSON file")
                sys.exit()

        result.prompt_for_missing_values()

        return result

    def prompt_for_missing_values(self):
        if self.username is None:
            self.username = input("Enter Username :")
        if self.password is None:
            self.password = getpass.getpass(f"Password for '{self.username}':")


import os

from pyega3.libs.data_file import DataFile


def test_in_some_folder(mock_data_client):
    folder = "FOO"
    file_id = "EGAF001"
    base_name = "filename"
    base_ext = ".ext"
    full_name = "/really/long/" + base_name + base_ext
    file = DataFile(mock_data_client, file_id, display_file_name=full_name)
    assert (
            file.generate_output_filename(folder, None)
            == os.path.join(folder, file_id, base_name + base_ext)
    )


def test_in_current_directory(mock_data_client):
    folder = os.getcwd()
    file_id = "EGAF001"
    base_name = "filename"
    base_ext = ".ext"
    full_name = "/really/long/" + base_name + base_ext
    file = DataFile(mock_data_client, file_id, display_file_name=full_name)
    assert (
            file.generate_output_filename(folder, None)
            == os.path.join(folder, file_id, base_name + base_ext)
    )


def test_with_genomic_range(mock_data_client):
    folder = os.getcwd()
    file_id = "EGAF001"
    base_name = "filename"
    base_ext = ".ext"
    full_name = "/really/long/" + base_name + base_ext
    file = DataFile(mock_data_client, file_id, display_file_name=full_name)
    assert (
            file.generate_output_filename(folder, ("chr1", None, 100, 200, "CRAM"))
            == os.path.join(folder, file_id, base_name + "_genomic_range_chr1_100_200" + base_ext + ".cram")
    )



import hashlib
import logging
import logging.handlers
import os
import time

import requests
from tqdm import tqdm


def merge_bin_files_on_disk(target_file_name, files_to_merge, downloaded_file_total_size):
    logging.info('Combining file chunks (this operation can take a long time depending on the file size)')
    start = time.time()

    with tqdm(total=int(downloaded_file_total_size), unit='B', unit_scale=True) as pbar:
        os.rename(files_to_merge[0], target_file_name)
        logging.debug(files_to_merge[0])
        if pbar:
            pbar.update(os.path.getsize(target_file_name))

        with open(target_file_name, 'ab') as target_file:
            for file_name in files_to_merge[1:]:
                with open(file_name, 'rb') as f:
                    logging.debug(file_name)
                    copyfileobj(f, target_file, 65536, pbar)
                os.remove(file_name)
        pbar.close()

    end = time.time()
    logging.debug(f'Merged in {end - start} sec')


def copyfileobj(f_source, f_destination, length=16 * 1024, pbar=None):
    while 1:
        buf = f_source.read(length)
        if not buf:
            break
        f_destination.write(buf)
        pbar.update(len(buf))


def calculate_md5(fname, file_size):
    if not os.path.exists(fname):
        raise Exception(f"Local file '{fname}' does not exist")
    hash_md5 = hashlib.md5()

    with tqdm(total=int(file_size), unit='B', unit_scale=True) as pbar:
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
                if pbar:
                    pbar.update(len(chunk))
        pbar.close()
    return hash_md5.hexdigest()


def get_fname_md5(fname):
    return fname + ".md5"


def md5(fname, file_size):
    fname_md5 = get_fname_md5(fname)
    # check if md5 has been previously stored in aux file
    if os.path.exists(fname_md5) and os.path.getsize(fname_md5) == 32:
        logging.info(f"Reusing pre-calculated md5: '{fname_md5}'")
        with open(fname_md5, 'rb') as f:
            return f.read().decode()
    # now do the real calculation
    result = calculate_md5(fname, file_size)
    return result


def status_ok(status_string):
    if status_string == "available":
        return True
    else:
        return False


def get_client_ip():
    endpoint = 'https://ipinfo.io/json'
    unknown_status = 'Unknown'
    try:
        response = requests.get(endpoint, verify=True)
        if response.status_code != 200:
            print('Status:', response.status_code, 'Problem with the request.')
            return unknown_status

        data = response.json()
        return data['ip']
    except Exception:
        logging.error("Failed to obtain IP address")
        return unknown_status


def verify_output_dir(output_dir):
    """
    Checks whether the directory, specified by the "output_dir" parameter,
    exists or not. If "output_dir" points to a non-existent directory,
    then a NotADirectoryError exception is thrown, otherwise the absolute path
    of that directory is returned.
    """
    absolut_path_of_output_dir = os.path.abspath(output_dir)

    if os.path.isdir(absolut_path_of_output_dir):
        return absolut_path_of_output_dir
    else:
        raise NotADirectoryError(f'The "{output_dir}" directory, which was specified by the --output-dir '
                                 'command-line argument, is not an existing directory. '
                                 'Please either create that directory or specify a different one.')



import sys

from setuptools import setup, find_packages

VERSION = "4.0.4"

CURRENT_PYTHON_VERSION = sys.version_info[:2]

# When changing the value of the REQUIRED_PYTHON_VERSION variable,
# make sure to also change the "python_requires" variable
# and the "classifiers" section in this file (setup.py).
REQUIRED_PYTHON_VERSION = (3, 6)

if CURRENT_PYTHON_VERSION < REQUIRED_PYTHON_VERSION:
    sys.stderr.write("""
==========================
Unsupported Python version
==========================
This version of pyEGA3 requires Python {}.{}, but you're trying to
install it on Python {}.{}. Please try to upgrade to a newer Python
version or maybe use pyEGA3 in a Docker container (see the README for that).
""".format(*(REQUIRED_PYTHON_VERSION + CURRENT_PYTHON_VERSION)))
    sys.exit(1)

with open("README.md", encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="pyega3",
    description="EGA python client",
    long_description=long_description,
    long_description_content_type="text/plain",
    packages=find_packages(),
    version=VERSION,
    author="EGA team",
    author_email="ega-helpdesk@ebi.ac.uk",
    python_requires=">=3.6",
    install_requires=["requests==2.26.0", "tqdm==4.19.6", "htsget==0.2.5", "psutil==5.6.6", "urllib3==1.26.7"],
    tests_require=["pytest~=6.2.5", "coverage==4.5.1", "responses~=0.16.0", "pyfakefs~=4.5.3"],
    keywords=["EGA", "archive"],
    license="Apache License, Version 2.0",
    url="https://github.com/EGA-archive/ega-download-client",
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6"
    ],
    entry_points={
        "console_scripts": [
            "pyega3 = pyega3.pyega3:main",
        ]
    }
)



import unittest.mock
from unittest import mock

import pytest

from pyega3.libs.data_file import DataFile
from pyega3.libs.data_set import DataSet


@mock.patch("pyega3.libs.data_file.DataFile")
def test_calls_download_for_every_file_in_dataset(mocked_datafile, mock_data_server, dataset_with_files,
                                                  mock_server_config,
                                                  mock_auth_client, mock_data_client):
    num_connections = 5

    mock_files = [unittest.mock.MagicMock() for _ in range(len(dataset_with_files.files))]
    for file in mock_files:
        file.status = "available"
    mocked_datafile.side_effect = mock_files

    dataset = DataSet(mock_data_client, dataset_with_files.id)
    dataset.download(num_connections, None, None, 5, 5)

    assert len(dataset_with_files.files) == mocked_datafile.call_count
    mocked_datafile.assert_has_calls(
        [mock.call(mock_data_client,
                   f['fileId'],
                   display_file_name=f['displayFileName'],
                   file_name=f['fileName'],
                   size=f['fileSize'],
                   unencrypted_checksum=f['unencryptedChecksum'],
                   status=f['fileStatus'])
         for f in dataset_with_files.files])

    for mock_file in mock_files:
        assert len(mock_file.method_calls) == 1
        assert mock_file.method_calls[0] == ('download_file_retry',
                                             (num_connections, None, None, 5, 5, DataFile.DEFAULT_SLICE_SIZE))


@mock.patch("pyega3.libs.data_file.DataFile")
def test_only_download_available_files(mocked_datafile, mock_server_config, mock_data_server, dataset_with_files,
                                       mock_auth_client, mock_data_client):
    num_connections = 5

    mock_files = [unittest.mock.MagicMock() for _ in range(len(dataset_with_files.files))]
    for file in mock_files:
        file.status = "available"
    mocked_datafile.side_effect = mock_files

    mock_data_server.files[dataset_with_files.files[0].get("fileId")] = dataset_with_files.files[0] = {
        "fileId": dataset_with_files.files[0].get("fileId"),
        "fileStatus": "not available"
    }
    mock_files[0].status = "not available"

    dataset = DataSet(mock_data_client, dataset_with_files.id)
    dataset.download(num_connections, None, None, 5, 5)

    assert len(dataset_with_files.files) == mocked_datafile.call_count
    mocked_datafile.assert_has_calls(
        [mock.call(mock_data_client, f['fileId'],
                   display_file_name=f.get('displayFileName'),
                   file_name=f.get('fileName'),
                   size=f.get('fileSize'),
                   unencrypted_checksum=f.get('unencryptedChecksum'),
                   status=f.get('fileStatus')) for f in dataset_with_files.files])

    # The first file was not available so it should not have been called
    assert len(mock_files[0].method_calls) == 0

    # The other files should all have been downloaded
    for mock_file in mock_files[1:]:
        assert len(mock_file.method_calls) == 1
        assert mock_file.method_calls[0] == ('download_file_retry',
                                             (num_connections, None, None, 5, 5, DataFile.DEFAULT_SLICE_SIZE))


def test_no_error_if_md5_mismatch(mock_server_config, mock_data_server, dataset_with_files, mock_auth_client,
                                  mock_data_client):
    def dfr_throws(p1, p2, p3, p4, p5, p6): raise Exception("bad MD5")

    with mock.patch("pyega3.libs.data_file.DataFile.download_file_retry", dfr_throws):
        dataset = DataSet(mock_data_client, dataset_with_files.id)
        dataset.download(1, None, None, 5, 5)


def test_download_unknown_dataset_does_not_call_download_file_retry(mock_server_config, mock_data_server,
                                                                    mock_auth_client, mock_data_client):
    with mock.patch("pyega3.libs.data_file.DataFile.download_file_retry") as mocked_dfr:
        bad_dataset = "EGAD00000000666"
        dataset = DataSet(mock_data_client, bad_dataset)
        dataset.download(1, None, None, 5, 5)
        assert 0 == mocked_dfr.call_count


def test_download_legacy_dataset(mock_server_config, mock_auth_client, mock_data_client):
    with pytest.raises(SystemExit):
        dataset = DataSet(mock_data_client, "EGAD00000000003")
        dataset.download(1, None, None, 5, 5)



import pytest
import requests

import tests.unit.conftest as common
from pyega3.libs.data_set import DataSet


def test_list_all_datasets(mock_data_server, mock_data_client):
    mock_data_server.dataset_files = {'EGAD000000001': None, 'EGAD000000002': None, 'EGAD000000003': None}
    resp_json = DataSet.list_authorized_datasets(mock_data_client)
    assert len(resp_json) == 3
    assert resp_json[0].id == mock_data_server.all_datasets[0]
    assert resp_json[1].id == mock_data_server.all_datasets[1]
    assert resp_json[2].id == mock_data_server.all_datasets[2]


def test_error_when_token_is_not_valid(mock_data_server, mock_data_client):
    mock_data_client.auth_client = common.MockAuthClient("invalid token")
    with pytest.raises(requests.exceptions.HTTPError):
        DataSet.list_authorized_datasets(mock_data_client)


def test_exit_when_user_has_no_datasets(mock_data_server, mock_data_client):
    mock_data_server.dataset_files = None

    with pytest.raises(SystemExit):
        DataSet.list_authorized_datasets(mock_data_client)



import pytest
import requests

import tests.unit.conftest as common
from pyega3.libs.data_set import DataSet


def test_list_all_files_in_dataset(dataset_with_files, mock_data_client):
    resp_json = DataSet(mock_data_client, dataset_with_files.id).list_files()

    assert len(resp_json) == 2
    assert resp_json[0].id == dataset_with_files.files[0]['fileId']
    assert resp_json[1].id == dataset_with_files.files[1]['fileId']


def test_error_with_bad_token(mock_data_server, dataset_with_files, mock_server_config, mock_data_client):
    mock_data_client.auth_client = common.MockAuthClient(common.rand_str())
    with pytest.raises(requests.exceptions.HTTPError):
        DataSet(mock_data_client, dataset_with_files.id).list_files()


def test_exit_with_unknown_dataset(mock_data_client, dataset_with_files):
    bad_dataset = common.rand_str()
    with pytest.raises(SystemExit):
        DataSet(mock_data_client, bad_dataset).list_files()


def test_exit_with_legacy_dataset(mock_data_client):
    with pytest.raises(SystemExit):
        DataSet(mock_data_client, "EGAD00000000003").list_files()


def test_exit_when_no_file_in_dataset(empty_dataset, mock_data_client):
    with pytest.raises(SystemExit):
        DataSet(mock_data_client, empty_dataset.id).list_files()
